from typing import Union
from previsionio.usecase_version import ClassicUsecaseVersion
import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score, fbeta_score, recall_score, precision_score


def cv_classif_analysis(usecase: ClassicUsecaseVersion, thresh: float = None, step: Union[int, float] = 1000):
    '''Get metrics on a CV file retrieved from the platform for a binary classification usecase

    Args:
        usecase (Usecase): usecase to analyze
        thresh (float, optional): threshold to use (if none is provided, optimal threshold will be computed
            given F1 score)
        step (int): number of iterations required to find optimal threshold (1000 by default = 0.1% resolution per fold)
    Returns:
        metrics computed for the CV (Dataframe)
    '''
    step = float(step)

    # get cv
    cv = usecase.get_cv()

    if '__fold__' not in cv.columns:
        raise Exception('The fold column is missing in the cv dataframe')

    fold = cv['__fold__'].unique()
    AUC = []
    f1 = []
    f1_fold = []
    precision_fold = []
    recall_fold = []
    threshold = []

    target_col_name = usecase.column_config.target_column
    assert target_col_name
    pred_target_col_name = 'pred_' + target_col_name

    for i in sorted(fold):
        f = cv['__fold__'] == i
        actual = cv[target_col_name][f].values
        predicted = cv[pred_target_col_name][f].values

        # compute auc per fold
        AUC.append(roc_auc_score(actual, predicted))

        # find optimal threshold for F1 score per fold
        f1 = []
        for j in range(1, int(step) + 1):
            idx = predicted >= j / step
            p = np.zeros(len(predicted))
            p[idx] = 1
            f1.append((j, fbeta_score(actual, p, beta=1)))
        f1 = pd.DataFrame(f1, columns=['j', 'f1'])

        # best threshold is the one that maximises f1 (if not provided by user)
        if thresh is None:
            threshold.append(f1['j'][f1['f1'].idxmax()] / step)
        else:
            threshold = np.repeat(thresh, len(fold))

        # store f1 score/recall/precision/specifity for optimal thresh per fold
        f1_fold.append(fbeta_score(cv[target_col_name][f].values,
                                   (cv[pred_target_col_name][f] >= threshold[i - 1]).astype('int'),
                                   beta=1))
        recall_fold.append(recall_score(cv[target_col_name][f].values,
                                        (cv[pred_target_col_name][f] >= threshold[i - 1]).astype('int')))
        precision_fold.append(precision_score(cv[target_col_name][f].values,
                                              (cv[pred_target_col_name][f] >= threshold[i - 1]).astype('int')))

    print('F1 by fold:', f1_fold)
    print('Precision by fold:', precision_fold)
    print('Recall by fold:', recall_fold)
    print('Threshold by fold:', threshold)

    return {
        'auc': np.round(np.mean(AUC), 4),
        'f1': np.round(np.mean(f1_fold), 4),
        'precision': np.round(np.mean(precision_fold), 4),
        'recall': np.round(np.mean(recall_fold), 4),
        'threshold': np.round(np.mean(threshold), 4)
    }



import copy
from enum import Enum
from typing import List, Union


def _drop_in_list(feature_list: List, param_to_drop: List):
    new = copy.copy(feature_list)
    for arg in param_to_drop:
        new.remove(arg.value)
    return new


class TypeProblem(Enum):
    """ Type of supervised problems available with Prevision.io. """
    Classification = 'classification'
    """Prediction using classification approach, for when the output variable is a category"""
    Regression = 'regression'
    """Prediction using regression problem, for when the output variable is a real or continuous value"""
    MultiClassification = 'multiclassification'
    """Prediction using classification approach, for when the output variable many categories"""
    TextSimilarity = 'text-similarity'
    """Ranking of texts by keywords"""
    ObjectDetection = "object-detection"
    """Detection of pattern in images"""


class DataType(Enum):
    """ Type of data available with Prevision.io. """
    Tabular = 'tabular'
    """Data arranged in a table"""
    TimeSeries = 'timeseries'
    """Data points indexed in time order"""
    Images = 'images'
    """Catalogue of images"""


class UsecaseState(Enum):
    """ Possible state of a Usecase in Prevision.io """
    Done = 'done'
    """The usecase finished properly"""
    Running = 'running'
    """The usecase is still running"""
    Failed = 'failed'
    """The usecase finished with an error"""
    Pending = 'pending'
    """The usecase is waiting for hardware ressources"""


class AdvancedModel(Enum):
    """ Types of normal models that can be trained with Prevision.io.
    The ``Full`` member is a shortcut to get all available models at once.
    To just drop a single model from a list of models, use:

    .. code-block:: python

        LiteModel.drop(LiteModel.xxx)
    """
    LightGBM = 'LGB'
    """LightGBM"""
    XGBoost = 'XGB'
    """XGBoost"""
    NeuralNet = 'NN'
    """NeuralNet"""
    ExtraTrees = 'ET'
    """ExtraTrees"""
    LinReg = 'LR'
    """Linear Regression"""
    RandomForest = 'RF'
    """Random Forest"""
    CatBoost = 'CB'
    """CatBoost"""
    Full = [LightGBM, XGBoost, NeuralNet, ExtraTrees, LinReg, RandomForest, CatBoost]
    """Evaluate all models"""


class NormalModel(Enum):
    """ Types of lite models that can be trained with Prevision.io.
    The ``Full`` member is a shortcut to get all available models at once.
    To just drop a single model from a list of models, use:

    .. code-block:: python

        Model.drop(Model.xxx)
    """
    LightGBM = 'LGB'
    """LightGBM"""
    XGBoost = 'XGB'
    """XGBoost"""
    NeuralNet = 'NN'
    """NeuralNet"""
    ExtraTrees = 'ET'
    """ExtraTrees"""
    LinReg = 'LR'
    """Linear Regression"""
    RandomForest = 'RF'
    """Random Forest"""
    NaiveBayesClassifier = 'NBC'
    """Random Forest"""
    CatBoost = 'CB'
    """CatBoost"""
    Full = [LightGBM, XGBoost, NeuralNet, ExtraTrees, LinReg, RandomForest, NaiveBayesClassifier, CatBoost]
    """Evaluate all models"""


class SimpleModel(Enum):
    """ Types of simple models that can be trained with Prevision.io.
    The ``Full`` member is a shortcut to get all available simple models at once.
    To just drop a single model from a list of simple models, use:

    .. code-block:: python

        SimpleModel.drop(SimpleModel.xxx)
    """
    DecisionTree = 'DT'
    """DecisionTree"""
    LinReg = 'LR'
    """Linear Regression"""
    Full = [DecisionTree, LinReg]
    """Evaluate all simple models"""


class Feature(Enum):
    """ Types of feature engineering that can be applied to a dataset with Prevision.io.
    The ``Full`` member is a shortcut to get all available feature engineering modules at once.
    To just drop a feature engineering module from a list of modules, use:

    .. code-block:: python

        Feature.drop(Feature.xxx)
    """
    Counts = 'Counter'
    """Value type counting"""
    DateTime = 'Date'
    """Date transformation"""
    Frequency = 'freq'
    """Frequency encoding"""
    TextTfidf = 'text_tfidf'
    """Statistical analysis"""
    TextWord2vect = 'text_word2vec'
    """Word embedding"""
    TextEmbedding = 'text_embedding'
    """Sentence embedding"""
    TargetEncoding = 'tenc'
    """Target encoding"""
    PolynomialFeatures = 'poly'
    """Polynomial feature"""
    PCA = 'pca'
    """Principal component analysis"""
    KMeans = 'kmean'
    """K-Means clustering"""
    Full = [Counts, DateTime, Frequency, TextTfidf,
            TextWord2vect, TextEmbedding, TargetEncoding,
            PolynomialFeatures, PCA, KMeans]
    """Full feature engineering"""


class Profile(Enum):
    """ Training profile type. """
    Quick = 'quick'
    """Quickest profile, lowest predictive performance"""
    Normal = 'normal'
    """Normal profile, best balance"""
    Advanced = 'advanced'
    """Slowest profile, for maximal optimization"""


class UsecaseConfig(object):

    config = {}

    @classmethod
    def from_dict(cls, kwargs_dict):
        reverse_dict = dict([(v, k) for k, v in cls.config.items()])

        class_args = {}

        for k, v in kwargs_dict.items():
            if k not in reverse_dict:
                continue

            if v == '' or not v:
                continue

            class_args[reverse_dict[k]] = v

        return cls(**class_args)


class TrainingConfig(UsecaseConfig):
    """ Training configuration that holds the relevant data for a usecase description:
    the wanted feature engineering, the selected models, the training speed...

    Args:
        profile (Profile): Type of training profile to use:

            - "quick": this profile runs very fast but has a lower performance
              (it is recommended for early trials)
            - "advanced": this profile runs slower but has increased performance
              (it is usually for optimization steps at the end of your project)
            - the "normal" profile is something in-between to help you investigate
              an interesting result

        advanced_models (list(AdvancedModel), optional): Names of the advanced models to use in the usecase
            (among: "LR", "RF", "ET", "XGB", "LGB", "CB" and "NN"). The advanced models will be
            hyperparametrized, resulting in a more accurate modelization at the cost of a longer
            training time.
        normal_models (list(NormalModel), optional): Names of the (normal) models to use in the usecase
            (among: "LR", "RF", "ET", "XGB", "LGB", "CB", 'NB' and "NN"). The normal models only
            use default parameters.
        simple_models (list(SimpleModel), optional): Names of the (simple) models to use in the usecase
            (among: "LR" and "DT"). These models are easy to ineterpret and fast to train but only
            offer a limited modelization complexity.
        features (list(Feature), optional): Names of the feature engineering modules to use (among:
            "Counter", "Date", "freq", "text_tfidf", "text_word2vec", "text_embedding", "tenc",
            "ee", "poly", "pca" and "kmean")
        with_blend (bool, optional): If true, Prevision.io's pipeline will add "blend" models
            at the end of the training by cherry-picking already trained models and fine-tuning
            hyperparameters (usually gives even better performance)
        feature_time_seconds (int, optional): feature selection take at most fsel_time in seconds
        feature_number_kept (int, optional): a feature selection algorithm is launched to keep at most
            `feature_number_kept` features
    """

    config = {
        'features': 'features_engineering_selected_list',
        'feature_time_seconds': 'features_selection_time',
        'feature_number_kept': 'features_selection_count',
        'advanced_models': 'normal_models',
        'normal_models': 'lite_models'
    }

    def __init__(self,
                 profile: Profile = Profile.Quick,
                 advanced_models: List[AdvancedModel] = [AdvancedModel.XGBoost, AdvancedModel.LinReg],
                 normal_models: List[NormalModel] = [NormalModel.XGBoost, NormalModel.LinReg],
                 simple_models: List[SimpleModel] = [],
                 features: List[Feature] = [Feature.Frequency, Feature.TargetEncoding, Feature.Counts],
                 with_blend: bool = False,
                 feature_time_seconds: int = 3600,
                 feature_number_kept: Union[int, None] = None):
        """

        Args:
            profile:
            normal_models:
            lite_models:
            simple_models:
            features (Feature, optional): Names of the feature engineering modules to use
            with_blend (bool, optional): models selectioned are also launched as blend
            feature_time_seconds (int, optional): feature selection take at most fsel_time in seconds
            feature_number_kept (int, optional): a feature selection algorithm is launched to keep at most
                `feature_number_kept` features
        """

        self.features = features

        self.advanced_models = advanced_models
        self.normal_models = normal_models
        self.simple_models = simple_models

        self.profile = profile
        self.with_blend = with_blend
        self.feature_time_seconds = feature_time_seconds

        if feature_number_kept:
            self.feature_number_kept = feature_number_kept


class YesOrNo(Enum):
    Yes = "yes"
    No = "no"


class YesOrNoOrAuto(Enum):
    Yes = "yes"
    No = "no"
    Auto = "auto"


class ColumnConfig(UsecaseConfig):
    """ Column configuration for starting a usecase: this object defines
    the role of specific columns in the dataset (and optionally the list of columns
    to drop).

    Args:
        target_column (str, optional): Name of the target column in the dataset
        id_column (str, optional): Name of the id column in the dataset that does
            not have any signal and will be ignored for computation
        fold_column (str, optional): Name of the fold column used that should be used to
            compute the various folds in the dataset
        weight_column (str, optional): Name of the weight column used to assign non-equal
            importance weights to the various rows in the dataset
        filename_column (str, optional): Name of the filename column in the dataset for
            an image-based usecase
        time_column (str, optional): Name of the time column in the dataset for a
            timeseries usecase
        group_columns (list(str), optional): Names of the columns in the dataset that define a
            unique time serie for a timeseries usecase
        apriori_columns (list(str), optional): Names of the columns that are known *a priori* in
            the dataset for a timeseries usecase
        drop_list (list(str), optional): Names of all the columns that should be dropped
            from the dataset while training the usecase
    """

    config = {
        'group_columns': 'group_list',
        'apriori_columns': 'apriori_list',
        'drop_list': 'drop_list',
    }

    def __init__(self,
                 target_column: Union[str, None] = None,
                 filename_column: Union[str, None] = None,
                 id_column: Union[str, None] = None,
                 fold_column: Union[str, None] = None,
                 weight_column: Union[str, None] = None,
                 time_column: Union[str, None] = None,
                 group_columns: Union[List[str], None] = None,
                 apriori_columns: Union[List[str], None] = None,
                 drop_list: Union[List[str], None] = None):
        self.target_column: Union[str, None] = target_column
        self.filename_column: Union[str, None] = filename_column
        self.id_column: Union[str, None] = id_column
        self.fold_column: Union[str, None] = fold_column
        self.weight_column: Union[str, None] = weight_column
        self.time_column: Union[str, None] = time_column
        self.group_columns: Union[List[str], None] = group_columns
        self.apriori_columns: Union[List[str], None] = apriori_columns
        self.drop_list: Union[List[str], None] = drop_list


base_config = TrainingConfig(profile=Profile.Normal,
                             advanced_models=AdvancedModel.Full.value,
                             normal_models=NormalModel.Full.value,
                             simple_models=SimpleModel.Full.value,
                             features=_drop_in_list(Feature.Full.value, [Feature.PCA, Feature.KMeans]),
                             with_blend=True)

quick_config = TrainingConfig(profile=Profile.Quick,
                              advanced_models=_drop_in_list(AdvancedModel.Full.value, [AdvancedModel.NeuralNet]),
                              normal_models=_drop_in_list(NormalModel.Full.value, [NormalModel.NeuralNet]),
                              simple_models=_drop_in_list(SimpleModel.Full.value, [SimpleModel.LinReg]),
                              features=_drop_in_list(Feature.Full.value, [Feature.PCA, Feature.KMeans]),
                              with_blend=False)

ultra_config = TrainingConfig(profile=Profile.Quick,
                              features=_drop_in_list(Feature.Full.value, [Feature.PCA, Feature.KMeans]),
                              advanced_models=[AdvancedModel.XGBoost],
                              normal_models=[NormalModel.XGBoost],
                              simple_models=[SimpleModel.LinReg],
                              with_blend=False)

nano_config = TrainingConfig(profile=Profile.Quick,
                             advanced_models=[AdvancedModel.LinReg],
                             normal_models=[NormalModel.LinReg],
                             simple_models=[],
                             features=[],
                             with_blend=False)



from typing import Union
import requests
from . import client
from .utils import parse_json
from .api_resource import ApiResource, UniqueResourceMixin


class Connector(ApiResource, UniqueResourceMixin):
    """ A connector to interact with a distant source of data (and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        conn_type (str): Type of the connector, among "FTP", "SFTP", "SQL", "S3",
            "HIVE", "HBASE", "GCP"
        username (str, optional): Username to use connect to the remote data source
        password (str, optional): Password to use connect to the remote data source
    """

    resource = 'connectors'
    conn_type = 'connector'

    def __init__(self, _id: str, name: str, host: str = None, port: int = None, type: str = None,
                 username: str = '', password: str = '', googleCredentials: str = None, **kwargs):
        super().__init__(_id=_id, name=name, host=host, port=port, conn_type=type,
                         username=username, password=password, googleCredentials=googleCredentials)
        self._id = _id
        self.name = name
        self.host = host
        self.port = port
        self.type = type
        self.username = username
        self.password = password
        self.googleCredentials = googleCredentials

        self.other_params = kwargs

    @classmethod
    def list(cls, project_id: str, all: bool = False):
        """ List all the available connectors in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Connector` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Connector`): Fetched connector objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**conn_data) for conn_data in resources
                if conn_data['type'] == cls.conn_type or cls.conn_type == 'connector']

    @classmethod
    def _new(cls, project_id: str, name: str, host: str, port: Union[int, None], conn_type: str, username: str = None,
             password: str = None, googleCredentials: str = None):
        """ Create a new connector object on the platform.

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            conn_type (str): Type of the connector, among "FTP", "SFTP", "SQL", "S3",
                "HIVE", "HBASE" or "GCP"
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source

        Returns:
            :class:`.Connector`: Newly create connector object
        """
        data = {
            'name': name,
            'host': host,
            'port': port,
            'type': conn_type
        }
        message_prefix = 'New connector creation'
        if username:
            data['username'] = username
        if password:
            data['password'] = password
        if googleCredentials:
            data['googleCredentials'] = googleCredentials
            content_type = 'application/json'
            resp = client.request('/projects/{}/{}'.format(project_id, cls.resource), data=data, method=requests.post,
                                  content_type=content_type, message_prefix=message_prefix)
        else:
            resp = client.request('/projects/{}/{}'.format(project_id, cls.resource), data=data, method=requests.post,
                                  message_prefix=message_prefix)

        resp_json = parse_json(resp)
        if '_id' not in resp_json:
            if 'message' in resp_json:
                raise Exception('Prevision.io error: {}'.format(' '.join(resp_json['message'])))
            else:
                raise Exception('unknown error:{}'.format(resp_json))
        data['_id'] = resp_json['_id']

        return cls(**data)

    def test(self):
        """ Test a connector already uploaded on the platform.

        Returns:
            dict: Test results
        """
        resp = client.request('/connectors/{}/test'.format(self.id), method=requests.post, check_response=False)
        if resp.status_code == 200:
            return True
        else:
            return False


class DataTableBaseConnector(Connector):

    """ A specific type of connector to interact with a database client (containing databases and tables). """

    def list_databases(self):
        """ List all available databases for the client.

        Returns:
            dict: Databases information
        """
        url = '/{}/{}/databases'.format(self.resource, self._id)
        resp = client.request(url, requests.get, message_prefix='Databases listing')
        resp_json = parse_json(resp)
        return resp_json['items']

    def list_tables(self, database: str):
        """ List all available tables in a specific database for the client.

        Args:
            database (str): Name of the database to find tables for

        Returns:
            dict: Tables information
        """
        url = '/{}/{}/databases/{}/tables'.format(self.resource, self._id, database)
        resp = client.request(url, requests.get, message_prefix='Tables listing')
        resp_json = parse_json(resp)
        return resp_json['items']


class DataFileBaseConnector(Connector):
    """ A specific type of connector to interact with a database client (containing files). """

    def list_files(self):
        """ List all available tables in a specific database for the client.

        Args:
            database (str): Name of the database to find tables for

        Returns:
            dict: files information
        """
        url = '/{}/{}/paths'.format(self.resource, self._id)
        resp = client.request(url, requests.get, message_prefix='Datasource files listing')
        resp_json = parse_json(resp)
        return resp_json['items']


class FTPConnector(DataFileBaseConnector):

    """ A specific type of connector to interact with a FTP client (containing files). """

    conn_type = 'FTP'


class SFTPConnector(DataFileBaseConnector):

    """ A specific type of connector to interact with a secured FTP client (containing files). """

    conn_type = 'SFTP'


class SQLConnector(DataTableBaseConnector):

    """ A specific type of connector to interact with a SQL database client (containing databases and tables). """

    conn_type = 'SQL'


class HiveConnector(DataTableBaseConnector):

    """ A specific type of connector to interact with a Hive database client (containing databases and tables). """

    conn_type = 'HIVE'


# class HBaseConnector(DataBaseConnector):
#
#     """ A specific type of connector to interact with a HBase database client (containing databases and tables). """
#
#     conn_type = 'HBASE'
#
#     @classmethod
#     def new(cls, name, host, port=9090, username='', password=''):
#         return cls._new(name=name, host=host, conn_type='HBASE', port=port, username=username, password=password)


class S3Connector(Connector):

    """ A specific type of connector to interact with an Amazon S3 client (containing buckets with files). """

    conn_type = 'S3'


class GCPConnector(Connector):

    """ A specific type of connector to interact with a GCP database client
        (containing databases and tables or buckets)."""

    conn_type = 'GCP'

#
# class HDFSConnector(Connector):
#
#     """ A specific type of connector to interact with a HDFS client. """
#
#     conn_type = 'HDFS'
#
#     @classmethod
#     def new(cls, name, host, port=50070, username='', password=''):
#         return cls._new(name=name, host=host, conn_type='HDFS', port=port, username=username, password=password)


connectors_names = {
    'SQL': SQLConnector,
    'FTP': FTPConnector,
    'SFTP': SFTPConnector,
    'S3': S3Connector,
    'HIVE': HiveConnector,
    'GCP': GCPConnector
}



from typing import Dict
from requests.models import Response
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from previsionio.utils import NpEncoder
import json
from time import time
import requests
from . import logger
from . import config
from .utils import handle_error_response, parse_json, PrevisionException

PREVISION_TOKEN_URL = 'https://accounts.prevision.io/auth/realms/prevision.io/protocol/openid-connect/token'


class DeployedModel(object):

    """
    DeployedModel class to interact with a deployed model.

    Args:
        prevision_app_url (str): URL of the App. Can be retrieved on your app dashbord.
        client_id (str): Your app client id. Can be retrieved on your app dashbord.
        client_secret (str): Your app client secret. Can be retrieved on your app dashbord.
        prevision_token_url (str): URL of get token. Should be
            https://accounts.prevision.io/auth/realms/prevision.io/protocol/openid-connect/token
            if you're in the cloud, or a custom IP address if installed on-premise.
    """

    def __init__(self, prevision_app_url: str, client_id: str, client_secret: str, prevision_token_url: str = None):
        """Init DeployedModel (and check that the connection is valid)."""
        self.prevision_app_url = prevision_app_url
        self.client_id = client_id
        self.client_secret = client_secret
        if prevision_token_url:
            self.prevision_token_url = prevision_token_url
        else:
            self.prevision_token_url = PREVISION_TOKEN_URL
        self.problem_type = None

        self.token = None
        self.url = None

        self.token_creation_date = None
        self.access_token = None
        self.refresh_expires_in = None

        try:
            self._get_token()
            about_resp = self.request('/about', method=requests.get)
            app_info = parse_json(about_resp)
            self.problem_type = app_info['problem_type']
            inputs_resp = self.request('/inputs', method=requests.get)
            self.inputs = parse_json(inputs_resp)
            outputs_resp = self.request('/outputs', method=requests.get)
            self.outputs = parse_json(outputs_resp)
        except Exception as e:
            logger.error(e)
            raise PrevisionException('Cannot connect: {}'.format(e))

    def _generate_token(self):
        self.token_creation_date = time()
        client = BackendApplicationClient(client_id=self.client_id)
        oauth = OAuth2Session(client=client)
        token = oauth.fetch_token(token_url=self.prevision_token_url,
                                  client_id=self.client_id,
                                  client_secret=self.client_secret)
        self.token = token
        self.refresh_expires_in = token['refresh_expires_in']
        return token

    def _refresh_token(self):
        token = self.token
        extra = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }

        client = OAuth2Session(self.client_id, token=token)
        self.token = client.refresh_token(self.prevision_token_url, **extra)
        self.refresh_expires_in = self.token['refresh_expires_in']

        return self.token

    def _get_token(self):
        if self.token:
            if time() < self.token['expires_at'] - 20:
                return
            elif self.token_creation_date and self.refresh_expires_in:
                if time() < self.token_creation_date + self.refresh_expires_in:
                    # refresh token
                    self._refresh_token()
                else:
                    # generate token
                    self._generate_token()
            else:
                self._generate_token()
        else:
            self._generate_token()

    def check_types(self, features):
        for feature, value in features:
            pass

    def _check_token_url_app(self):

        if not self.prevision_app_url:
            raise PrevisionException('No url configured. Call client_app.init_client() to initialize')

        if not self.client_id:
            raise PrevisionException('No client id configured. Call client_app.init_client() to initialize')

        if not self.client_secret:
            raise PrevisionException('No client secret configured. Call client_app.init_client() to initialize')

    def request(self, endpoint, method, files=None, data=None, allow_redirects=True, content_type=None,
                check_response=True, message_prefix=None, **requests_kwargs):
        """
        Make a request on the desired endpoint with the specified method & data.

        Requires initialization.

        Args:
            endpoint: (str): api endpoint (e.g. /usecases, /prediction/file)
            method (requests.{get,post,delete}): requests method
            files (dict): files dict
            data (dict): for single predict
            content_type (str): force request content-type
            allow_redirects (bool): passed to requests method

        Returns:
            request response

        Raises:
            Exception: Error if url/token not configured
        """
        self._check_token_url_app()

        url = self.prevision_app_url + endpoint

        status_code = 502
        retries = config.request_retries
        n_tries = 0
        resp = None
        while (n_tries < retries) and (status_code in config.retry_codes):
            n_tries += 1

            try:
                self._get_token()
                assert self.token is not None
                headers = {
                    "Authorization": "Bearer " + self.token['access_token'],
                }
                if content_type:
                    headers['content-type'] = content_type

                resp = method(url,
                              headers=headers,
                              files=files,
                              allow_redirects=allow_redirects,
                              data=data,
                              **requests_kwargs)
                status_code = resp.status_code

            except Exception as e:
                logger.warning(f'Failed to request {url} retrying {retries - n_tries} times: {e.__repr__()}')
                if n_tries == retries:
                    raise PrevisionException(f'Error requesting: {url} after {n_tries} retries')
                continue

            if status_code in config.retry_codes:
                time.sleep(config.request_retry_time)

        assert isinstance(resp, Response)
        if check_response:
            handle_error_response(resp, url, data, message_prefix=message_prefix, n_tries=n_tries)

        return resp

    def predict(self, predict_data: Dict, use_confidence: bool = False, explain: bool = False):
        """ Get a prediction on a single instance using the best model of the usecase.

        Args:
            predict_data (dictionary): input data for prediction
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)
            explain (bool): Whether to explain prediction (default: ``False``)

        Returns:
            tuple(float, float, dict): Tuple containing the prediction value, confidence and explain.
            In case of regression problem type, confidence format is a list.
            In case of multiclassification problem type, prediction value format is a string.

        """
        # FIXME add some checks for feature name with input api
        features = [{'name': feature, 'value': value}
                    for feature, value in predict_data.items()]

        predict_url = '/predict'

        if explain or use_confidence:
            predict_url += '?'

        if explain:
            predict_url += 'explain=true&'

        if use_confidence:
            predict_url += 'confidence=true'

        predict_url = predict_url.rstrip('&')
        resp = self.request(predict_url,
                            data=json.dumps(features, cls=NpEncoder),
                            method=requests.post,
                            message_prefix='Deployed model predict')

        pred_response = resp.json()
        target_name = self.outputs[0]['keyName']
        preds = pred_response['response']['predictions']
        prediction = preds[target_name]
        if use_confidence:
            if self.problem_type == 'regression':
                confidance_resp = [{key: value} for key, value in preds.items() if 'TARGET_quantile=' in key]
            elif 'confidence' in preds:
                confidance_resp = preds['confidence']
            else:
                confidance_resp = None
        else:
            confidance_resp = None

        if explain and 'explanation' in preds:
            explain_resp = preds['explanation']
        else:
            explain_resp = None

        return prediction, confidance_resp, explain_resp



from enum import Enum


class Regression(Enum):
    """
    Metrics for regression projects
    Available metrics in prevision: rmse, mape, rmsle, mse, mae
    """
    RMSE = 'rmse'
    """Root Mean Squared Error"""
    RMSLE = 'rmsle'
    """Root Mean Squared Logarithmic Error"""
    MAPE = 'mape'
    """Mean Average Percentage Error"""
    MAE = 'mae'
    """Mean Average Error"""
    MSE = 'mse'
    """Mean squared Error"""
    RMSPE = 'rmspe'
    """Root Mean Squared Percentage Error"""
    MER = 'mer'
    """Median Absolute Error"""
    R2 = 'R2'
    """R2 Error"""
    SMAPE = 'smape'
    """Symmetric Mean Absolute Percentage Error"""


class Classification(Enum):
    """
    Metrics for classification projects
    Available metrics in prevision: auc, log_loss, error_rate_binary
    """
    AUC = 'auc'
    """Area Under ROC Curve"""
    log_loss = 'log_loss'
    """Logarithmic Loss"""
    error_rate = 'error_rate_binary'
    """Error rate"""
    accuracy = 'accuracy'
    """Accuracy"""
    F05 = 'F05'
    """F05 Score"""
    F1 = 'F1'
    """Balanced F-score"""
    F2 = 'F2'
    """F2 Score"""
    F3 = 'F3'
    """F3 Score"""
    F4 = 'F4'
    """F4 Score"""
    MCC = 'mcc'
    """Matthews correlation coefficient"""
    gini = 'gini'
    """Gini score"""
    AUCPR = 'aucpr'
    """precision recall area under the curve score"""
    Lift01 = 'lift_at_0.1'
    """lift at ratio 0.1"""
    Lift02 = 'lift_at_0.2'
    """lift at ratio 0.2"""
    Lift03 = 'lift_at_0.3'
    """lift at ratio 0.3"""
    Lift04 = 'lift_at_0.4'
    """lift at ratio 0.4"""
    Lift05 = 'lift_at_0.5'
    """lift at ratio 0.5"""
    Lift06 = 'lift_at_0.6'
    """lift at ratio 0.6"""
    Lift07 = 'lift_at_0.7'
    """lift at ratio 0.7"""
    Lift08 = 'lift_at_0.8'
    """lift at ratio 0.8"""
    Lift09 = 'lift_at_0.9'
    """lift at ratio 0.9"""


class MultiClassification(Enum):
    """
    Metrics for multiclassification projects
    """
    log_loss = 'log_loss'
    """Logarithmic Loss"""
    error_rate = 'error_rate_multi'
    """Multi-class Error rate"""
    macroF1 = 'macroF1'
    """balanced F-score"""
    AUC = 'auc'
    """Area Under ROC Curve"""
    accuracy = 'accuracy'
    """accuracy"""
    qkappa = 'qkappa'
    """quadratic weighted kappa"""
    MAP3 = 'map_at_3'
    """qmean average precision @3"""
    MAP5 = 'map_at_5'
    """qmean average precision @5"""
    MAP10 = 'map_at_10'
    """qmean average precision @10"""


class Clustering(Enum):
    """
    Metrics for clustering projects
    """
    silhouette = 'silhouette'
    """Clustering silhouette metric"""
    calinski_harabaz = 'calinski_harabaz'
    """Clustering calinski_harabaz metric"""


class TextSimilarity(Enum):
    """
    Metrics for text similarity projects
    """
    accuracy_at_k = 'accuracy_at_k'
    """Accuracy at K"""
    mrr_at_k = 'mrr_at_k'
    """Mean Reciprocal Rank at K"""



# -*- coding: utf-8 -*-
from __future__ import print_function
import json
from previsionio import metrics
from previsionio.model import Model
from typing import Dict, List, Union
import pandas as pd
import requests
import time
import previsionio as pio
import os
from functools import lru_cache
from dateutil import parser

from . import config
from .usecase_config import (AdvancedModel, DataType, Feature, NormalModel, Profile, SimpleModel,
                             TrainingConfig, ColumnConfig, TypeProblem, UsecaseState)
from .logger import logger
from .prevision_client import client
from .utils import parse_json, EventTuple, PrevisionException, zip_to_pandas, get_all_results
from .api_resource import ApiResource
from .dataset import Dataset
# from .usecase import Usecase


class BaseUsecaseVersion(ApiResource):

    """Base parent class for all usecases objects."""

    training_config: TrainingConfig
    column_config: ColumnConfig
    id_key = 'usecase_id'

    resource = 'usecase-versions'
    training_type: TypeProblem
    data_type: DataType
    model_class: Model

    def __init__(self, **usecase_info):
        super().__init__(**usecase_info)
        self.name: str = usecase_info.get('name', usecase_info['usecase'].get('name'))
        self._id: str = usecase_info['_id']
        self.usecase_id: str = usecase_info['usecase_id']
        self.project_id: str = usecase_info['project_id']
        self.dataset_id: str = usecase_info['dataset_id']
        self.holdout_dataset_id: Union[str, None] = usecase_info.get('holdout_dataset_id', None)
        self.created_at = parser.parse(usecase_info["created_at"])
        self._models = {}
        self.version = 1

    def __len__(self):
        return len(self.models)

    def update_status(self):
        return super().update_status(specific_url='/{}/{}'.format(self.resource,
                                                                  self._id))

    @classmethod
    def _from_id(cls, _id: str) -> Dict:
        """Get a usecase from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase to retrieve
            version (int, optional): Specific version of the usecase to retrieve
                (default: 1)

        Returns:
            :class:`.BaseUsecaseVersion`: Fetched usecase

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return super()._from_id(specific_url='/{}/{}'.format(cls.resource, _id))

    # @property
    # def usecase(self) -> 'Usecase':
    #     """Get a usecase of current usecase version.

    #     Returns:
    #         :class:`.Usecase`: Fetched usecase

    #     Raises:
    #         PrevisionException: Any error while fetching data from the platform
    #             or parsing result
    #     """
    #     return Usecase.from_id(self.usecase_id)

    @property
    def models(self):
        """Get the list of models generated for the current use case. Only the models that
        are done training are retrieved.

        Returns:
            list(:class:`.Model`): List of models found by the platform for the usecase
        """
        end_point = '/{}/{}/models'.format(self.resource, self._id)
        models = get_all_results(client, end_point, method=requests.get)
        for model in models:
            if model['_id'] not in self._models:
                self._models[model['_id']] = self.model_class(**model)
        return list(self._models.values())

    @property
    @lru_cache()
    def train_dataset(self):
        """ Get the :class:`.Dataset` object corresponding to the training dataset
        of the usecase.

        Returns:
            :class:`.Dataset`: Associated training dataset
        """
        return Dataset.from_id(_id=self.dataset_id)

    @property
    @lru_cache()
    def schema(self) -> dict:
        """ Get the data schema of the usecase.

        Returns:
            dict: Usecase schema
        """
        end_point = '/{}/{}/graph'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Usecase schema')
        uc_schema = json.loads(response.content.decode('utf-8'))
        return uc_schema

    @property
    def best_model(self):
        """ Get the model with the best predictive performance over all models (including
        Blend models), where the best performance corresponds to a minimal loss.

        Returns:
            (:class:`.Model`, None): Model with the best performance in the usecase, or
            ``None`` if no model matched the search filter.
        """
        best_model = None
        if len(self.models) == 0:
            raise PrevisionException('models not ready yet')
        for model in self.models:
            if 'best' in model.tags:
                best_model = model

        if best_model is None:
            best_model = self.models[0]
        return best_model

    @property
    def fastest_model(self):
        """Returns the model that predicts with the lowest response time

        Returns:
            Model object -- corresponding to the fastest model
        """
        fastest_model = None
        if len(self.models) == 0:
            raise PrevisionException('models not ready yet')
        for model in self.models:
            if 'fastest' in model.tags:
                fastest_model = model

        if fastest_model is None:
            fastest_model = self.models[0]

        return fastest_model

    @property
    def done(self) -> bool:
        """ Get a flag indicating whether or not the usecase is currently done.

        Returns:
            bool: done status
        """
        status = self._status
        return status['state'] == UsecaseState.Done.value

    @property
    def running(self) -> bool:
        """ Get a flag indicating whether or not the usecase is currently running.

        Returns:
            bool: Running status
        """
        status = self._status
        return status['state'] == UsecaseState.Running.value

    @property
    def status(self) -> UsecaseState:
        """ Get a flag indicating whether or not the usecase is currently running.

        Returns:
            bool: Running status
        """
        status = self._status
        return UsecaseState(status['state'])

    @property
    def advanced_models_list(self) -> List[AdvancedModel]:
        """ Get the list of selected advanced models in the usecase.

        Returns:
            list(AdvancedModel): Names of the normal models selected for the usecase
        """
        return [AdvancedModel(f) for f in self._status['usecase_version_params'].get('normal_models', [])]

    @property
    def normal_models_list(self) -> List[NormalModel]:
        """ Get the list of selected normal models in the usecase.

        Returns:
            list(NormalModel): Names of the normal models selected for the usecase
        """
        return [NormalModel(f) for f in self._status['usecase_version_params'].get('lite_models', [])]

    @property
    def simple_models_list(self) -> List[SimpleModel]:
        """ Get the list of selected simple models in the usecase.

        Returns:
            list(SimpleModel): Names of the simple models selected for the usecase
        """
        return [SimpleModel(f) for f in self._status['usecase_version_params'].get('simple_models', [])]

    def stop(self):
        """ Stop a usecase (stopping all nodes currently in progress). """
        logger.info('[Usecase] stopping usecase')
        end_point = '/{}/{}/stop'.format(self.resource, self._id)
        response = client.request(end_point,
                                  requests.put,
                                  message_prefix='Usecase stop')
        events_url = '/{}/{}'.format(self.resource, self._id)
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(self.resource_id,
                                                self.resource,
                                                EventTuple('USECASE_VERSION_UPDATE'),
                                                specific_url=events_url)
        logger.info('[Usecase] stopping:' + '  '.join(str(k) + ': ' + str(v)
                                                      for k, v in parse_json(response).items()))

    # def delete(self):
    #     """ Delete a usecase from the actual [client] workspace.
    #
    #     Returns:
    #         dict: Deletion process results
    #     """
    #     response = client.request(endpoint='/usecases/{}'.format(self._id),
    #                               method=requests.delete)
    #     return (json.loads(response.content.decode('utf-8')))

    def wait_until(self, condition, raise_on_error: bool = True, timeout: float = config.default_timeout):
        """ Wait until condition is fulfilled, then break.

        Args:
            condition (func: (:class:`.BaseUsecaseVersion`) -> bool.): Function to use to check the
                break condition
            raise_on_error (bool, optional): If true then the function will stop on error,
                otherwise it will continue waiting (default: ``True``)
            timeout (float, optional): Maximal amount of time to wait before forcing exit

        Example::

            usecase.wait_until(lambda usecasev: len(usecasev.models) > 3)

        Raises:
            PrevisionException: If the resource could not be fetched or there was a timeout.
        """
        t0 = time.time()
        while True:
            if timeout is not None and time.time() - t0 > timeout:
                raise PrevisionException('timeout while waiting on {}'.format(condition))

            try:
                if condition(self):
                    break
                elif self._status['state'] == 'failed':
                    raise PrevisionException('Resource failed while waiting')
            except PrevisionException as e:
                logger.warning(e.__repr__())
                if raise_on_error:
                    raise

            time.sleep(config.scheduler_refresh_rate)

    def get_holdout_predictions(self, full: bool = False):
        """
        Retrieves the list of holdout predictions for the current usecase from client workspace
        (with the full predictions object if necessary)
        Args:
            full (boolean): If true, return full holdout prediction objects (else only metadata)
        """
        end_point = '/usecase-versions/{}/holdout-predictions'.format(self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Holdout predictions listing')
        preds_list = (json.loads(response.content.decode('utf-8')))['items']
        preds_dict = {}
        for pred in preds_list:
            _id = pred.pop('_id')
            if full:
                end_point = '/predictions/{}/download'.format(_id)
                response = client.request(endpoint=end_point,
                                          method=requests.get,
                                          message_prefix='Predictions fetching')
                preds_dict[_id] = zip_to_pandas(response)
            else:
                preds_dict[_id] = pred
        return preds_dict

    def get_predictions(self, full: bool = False):
        """
        Retrieves the list of predictions for the current usecase from client workspace
        (with the full predictions object if necessary)
        Args:
            full (boolean): If true, return full prediction objects (else only metadata)
        """
        response = client.request(endpoint='/usecases/{}/predictions'.format(self._id),
                                  method=requests.get,
                                  message_prefix='Predictions listing')
        preds_list = (json.loads(response.content.decode('utf-8')))['items']
        preds_dict = {}
        for pred in preds_list:
            _id = pred.pop('_id')
            if full:
                response = client.request(endpoint='/predictions/{}/download'.format(_id),
                                          method=requests.get,
                                          message_prefix='Predictions fetching')
                preds_dict[_id] = zip_to_pandas(response)
            else:
                preds_dict[_id] = pred
        return preds_dict

    def delete_prediction(self, prediction_id: str):
        """ Delete a prediction in the list for the current usecase from the actual [client] workspace.

        Args:
            prediction_id (str): Unique id of the prediction to delete

        Returns:
            dict: Deletion process results
        """
        endpoint = '/usecases/{}/versions/{}/predictions/{}'.format(self._id, self.version, prediction_id)
        response = client.request(endpoint=endpoint,
                                  method=requests.delete,
                                  message_prefix='Prediction delete')
        return (json.loads(response.content.decode('utf-8')))

    def delete_predictions(self):
        """ Delete all predictions in the list for the current usecase from the actual [client] workspace.

        Returns:
            dict: Deletion process results
        """
        response = client.request(endpoint='/usecases/{}/versions/{}/predictions'.format(self._id, self.version),
                                  method=requests.delete,
                                  message_prefix='Predictions delete')
        return (json.loads(response.content.decode('utf-8')))

    @property
    def score(self) -> float:
        """ Get the current score of the usecase (i.e. the score of the model that is
        currently considered the best performance-wise for this usecase).

        Returns:
            float: Usecase score (or infinity if not available).
        """
        try:
            return self._status['score']
        except KeyError:
            return float('inf')

    def _save_json(self):
        raise NotImplementedError

    def save(self, directory: str = '.'):
        version_dict = self._save_json()
        with open(os.path.join(directory, self.name) + '.pio', 'w') as f:
            json.dump(version_dict, f)

    @classmethod
    def _load(cls, pio_file: str) -> Dict:
        with open(pio_file, 'r') as f:
            mdl = json.load(f)
        uc = cls._from_id(mdl['_id'])
        # TODO check holdout_dataset in usecase_version_params
        # if mdl['usecase_version_params'].get('holdout_dataset_id'):
        #     uc.holdout_dataset = mdl['usecase_version_params'].get('holdout_dataset_id')[0]
        return uc


class ClassicUsecaseVersion(BaseUsecaseVersion):

    def __init__(self, **usecase_info):
        super().__init__(**usecase_info)
        usecase_params = usecase_info['usecase_version_params']
        self.metric: str = usecase_params['metric']
        self.column_config = ColumnConfig(target_column=usecase_params.get('target_column'),
                                          fold_column=usecase_params.get('fold_column'),
                                          id_column=usecase_params.get('id_column'),
                                          weight_column=usecase_params.get('weight_column'),
                                          time_column=usecase_params.get('time_column', None),
                                          group_columns=usecase_params.get('group_columns'),
                                          apriori_columns=usecase_params.get('apriori_columns'),
                                          drop_list=usecase_params.get('drop_list'))

        self.training_config = TrainingConfig(profile=Profile(usecase_params.get('profile')),
                                              features=[Feature(f) for f in usecase_params.get(
                                                  'features_engineering_selected_list', [])],
                                              advanced_models=[
                                                  AdvancedModel(f) for f in usecase_params.get('normal_models', [])],
                                              normal_models=[NormalModel(f)
                                                             for f in usecase_params.get('lite_models', [])],
                                              simple_models=[SimpleModel(f)
                                                             for f in usecase_params.get('simple_models', [])],
                                              feature_time_seconds=usecase_params.get('features_selection_time', 3600),
                                              feature_number_kept=usecase_params.get('features_selection_count', None))

        self._id: str = usecase_info['_id']
        self.usecase_id: str = usecase_info['usecase_id']
        self.project_id: str = usecase_info['project_id']
        self.version = usecase_info.get('version', 1)
        self._usecase_info = usecase_info
        self.data_type: DataType = DataType(usecase_info['usecase'].get('data_type'))
        self.training_type: TypeProblem = TypeProblem(usecase_info['usecase'].get('training_type'))
        self.dataset_id: str = usecase_info['dataset_id']
        self.predictions = {}
        self.predict_token = None
        self._models = {}

    def print_info(self):
        """ Print all info on the usecase. """
        for k, v in self._usecase_info.items():
            print(str(k) + ': ' + str(v))

    @property
    @lru_cache()
    def correlation_matrix(self) -> pd.DataFrame:
        """ Get the correlation matrix of the features (those constitute the dataset
        on which the usecase was trained).

        Returns:
            ``pd.DataFrame``: Correlation matrix as a ``pandas`` dataframe
        """
        end_point = '/{}/{}/correlation-matrix'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Correlation matrix fetching')
        corr = json.loads(response.content.decode('utf-8'))
        var_names = [d['name'] for d in corr]
        matrix = pd.DataFrame(0, index=var_names, columns=var_names)
        for i, name in enumerate(var_names):
            matrix[name] = [d['score'] for d in corr[i]['correlation']]
        return matrix

    @property
    @lru_cache()
    def features(self) -> List[dict]:
        """ Get the general description of the usecase's features, such as:

        - feature types distribution
        - feature information list
        - list of dropped features

        Returns:
            dict: General features information
        """
        # todo pagination
        url = '/{}/{}/features'.format(self.resource, self._id)
        return get_all_results(client, url, method=requests.get)

    @property
    @lru_cache()
    def features_stats(self):
        """ Get the general description of the usecase's features, such as:

        - feature types distribution
        - feature information list
        - list of dropped features

        Returns:
            dict: General features information
        """
        # todo pagination
        end_point = '/{}/{}/features-stats'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Features stats fetching')
        return (json.loads(response.content.decode('utf-8')))

    @property
    @lru_cache()
    def dropped_features(self):
        """ Get dropped features

        Returns:
            dict: Dropped features
        """
        # todo pagination
        end_point = '/{}/{}/dropped-features'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Dropped features fetching')
        return (json.loads(response.content.decode('utf-8')))

    def get_feature_info(self, feature_name: str) -> Dict:
        """ Return some information about the given feature, such as:

        - name: the name of the feature as it was given in the ``feature_name`` parameter
        - type: linear, categorical, ordinal...
        - stats: some basic statistics such as number of missing values, (non missing) values
          count, plus additional information depending on the feature type:

            - for a linear feature: min, max, mean, std and median
            - for a categorical/textual feature: modalities/words frequencies,
              list of the most frequent tokens

        - role: whether or not the feature is a target/fold/weight or id feature (and for time
          series usecases, whether or not it is a group/apriori feature - check the
          `Prevision.io's timeseries documentation
          <https://previsionio.readthedocs.io/fr/latest/projects_timeseries.html>`_)
        - importance_value: scores reflecting the importance of the given feature

        Args:
            feature_name (str): Name of the feature to get informations about

            .. warning::

                The ``feature_name`` is case-sensitive, so "age" and "Age" are different
                features!

        Returns:
            dict: Dictionary containing the feature information

        Raises:
            PrevisionException: If the given feature name does not match any feaure
        """
        endpoint = '/{}/{}/features/{}'.format(self.resource, self._id, feature_name)
        response = client.request(endpoint=endpoint,
                                  method=requests.get,
                                  message_prefix='Features info fetching')

        result = (json.loads(response.content.decode('utf-8')))

        # drop chart-related informations
        keep_list = list(filter(lambda x: 'chart' not in x.lower(),
                                result.keys())
                         )
        return {k: result[k] for k in keep_list}

    @property
    def drop_list(self) -> List[str]:
        """ Get the list of drop columns in the usecase.

        Returns:
            list(str): Names of the columns dropped from the dataset
        """
        return self._status['usecase_version_params'].get('drop_list', [])

    @property
    def feature_list(self) -> List[Feature]:
        """ Get the list of selected feature engineering modules in the usecase.

        Returns:
            list(str): Names of the feature engineering modules selected for the usecase
        """
        res = [Feature(f) for f in self._status['usecase_version_params'].get('features_engineering_selected_list', [])]
        return res

    def get_cv(self) -> pd.DataFrame:
        """ Get the cross validation dataset from the best model of the usecase.

        Returns:
            ``pd.DataFrame``: Cross validation dataset
        """

        best = self.best_model

        return best.cross_validation

    @classmethod
    def _start_usecase(cls, project_id: str, name: str,
                       dataset_id: Union[str, List[str]], data_type: DataType, training_type: TypeProblem, **kwargs):
        """ Start a usecase of the given data type and problem type with a specific
        training configuration (on the platform).

        Args:
            name (str): Registration name for the usecase to create
            dataset_id (str|tuple(str, str)): Unique id of the training dataset resource or a tuple of csv and folder id
            data_type (str): Type of data used in the usecase (among "tabular", "images"
                and "timeseries")
            training_type: Type of problem to compute with the usecase (among "regression",
                "classification", "multiclassification" and "object-detection")
            **kwargs:

        Returns:
            :class:`.BaseUsecaseVersion`: Newly created usecase object
        """
        logger.info('[Usecase] Starting usecase')

        if data_type == DataType.Tabular or data_type == DataType.TimeSeries:
            data = dict(name=name, dataset_id=dataset_id, **kwargs)
        elif data_type == DataType.Images:
            csv_id, folder_id = dataset_id
            data = dict(name=name, dataset_id=csv_id, folder_dataset_id=folder_id, **kwargs)
        else:
            raise PrevisionException('invalid data type: {}'.format(data_type))
        endpoint = '/projects/{}/{}/{}/{}'.format(project_id, 'usecases', data_type.value, training_type.value)
        start = client.request(endpoint,
                               method=requests.post,
                               data=data,
                               content_type='application/json',
                               message_prefix='Usecase start')
        return parse_json(start)

    def predict_single(self,
                       data,
                       confidence=False,
                       explain=False):
        """ Get a prediction on a single instance using the best model of the usecase.

        Args:
            use_best_single (bool, optional): Whether to use the best single model
                instead of the best model overall (default: ``False``)
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)
            explain (bool): Whether to explain prediction (default: ``False``)

        Returns:
            dict: Dictionary containing the prediction.

            .. note::

                The format of the predictions dictionary depends on the problem type
                (regression, classification...)
        """

        best = self.best_model
        return best.predict_single(data,
                                   confidence=confidence,
                                   explain=explain)

    def predict_from_dataset(self,
                             dataset,
                             confidence=False,
                             dataset_folder=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the usecase.

        Arguments:
            dataset (:class:`.Dataset`): Reference to the dataset object to make
                predictions for
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)
            dataset_folder (:class:`.Dataset`): Matching folder dataset for the
                predictions, if necessary

        Returns:
            ``pd.DataFrame``: Predictions as a ``pandas`` dataframe
        """

        best = self.best_model

        return best.predict_from_dataset(dataset, confidence=confidence, dataset_folder=dataset_folder)

    def predict(self, df, confidence=False, prediction_dataset_name=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the usecase with a Scikit-learn style blocking prediction mode.

        .. warning::

            For large dataframes and complex (blend) models, this can be slow (up to 1-2 hours).
            Prefer using this for simple models and small dataframes, or use option ``use_best_single = True``.

        Args:
            df (``pd.DataFrame``): ``pandas`` DataFrame containing the test data
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)

        Returns:
            tuple(pd.DataFrame, str): Prediction data (as ``pandas`` dataframe) and prediction job ID.
        """

        best = self.best_model

        return best.predict(df=df, confidence=confidence, prediction_dataset_name=prediction_dataset_name)



from typing import Dict
import requests
from .utils import parse_json, get_all_results
from .prevision_client import client
from . import logger
from enum import Enum


class ApiResourceType(Enum):

    "All the different resource types and matching API endpoints."

    ApiResource = 'api_resource'
    Dataset = 'files/dataset'
    Folder = 'files/folder'
    Usecase = 'usecases'
    Connector = 'connectors'
    DataSource = 'datasources'
    Project = 'projects'


class UniqueResourceMixin:
    @classmethod
    def from_name(cls, name: str):
        resources = cls.list()
        resources_match = [c for c in resources if c.name == name]
        if resources_match:
            return resources_match[0]
        else:
            raise Exception('No such {}: {}'.format(cls.resource, name))


class ApiResource:

    """Base parent class for all SDK resource objects."""

    resource = ApiResourceType.ApiResource
    resource_params = []
    id_key = '_id'

    def __init__(self, **params):
        self._id: str = params.get('_id', "")
        if self._id == "":
            raise RuntimeError("Invalid _id received from {}".format(str(params)))
        self.resource_id = self._id
        # self.event_manager: Optional[EventManager] = None

    def update_status(self, specific_url: str = None) -> Dict:
        """Get an update on the status of a resource.

        Args:
            specific_url (str, optional): Specific (already parametrized) url to
                fetch the resource from (otherwise the url is built from the
                resource type and unique ``_id``)

        Returns:
            dict: Updated status info
        """
        if specific_url is None:
            url = '/{}/{}'.format(self.resource, self._id)
        else:
            url = specific_url
        # call api, add event to eventmanager events if available and return
        resource_status = client.request(url, method=requests.get,
                                         message_prefix='Update status {}'.format(self.resource))
        resource_status_dict = parse_json(resource_status)
        resource_status_dict['event_type'] = 'update'
        resource_status_dict['event_name'] = 'update'

        # if self.event_manager:
        #    self.event_manager.add_event(self.resource_id, resource_status_dict)

        return resource_status_dict

    @property
    def _status(self) -> Dict:

        # if self.event_manager:
        #     events = self.event_manager.events
        #     if self.resource_id in events:
        #         return sorted(events[self.resource_id],
        #                       key=lambda k: datetime.datetime.strptime(k['end'], '%Y-%m-%dT%H:%M:%S.%fZ'))[-1]
        #
        resource_status_dict = self.update_status()
        return resource_status_dict

    @property
    def id(self) -> str:
        return self._id

    def delete(self):
        """Delete a resource from the actual [client] workspace.

        Raises:
            PrevisionException: Any error while deleting data from the platform
        """
        _ = client.request('/{}/{}'.format(self.resource, self._id),
                           method=requests.delete,
                           message_prefix='Delete {}'.format(self.resource))
        logger.info('[Delete {} OK]'.format(self.resource))

    @classmethod
    def _from_id(cls, _id: str = None, specific_url: str = None) -> Dict:
        """Get a resource from the platform by its unique id.
        You must provide either an ``_id`` or a ``specific_url``.

        Args:
            _id (str, optional): Unique id of the resource to retrieve
            specific_url (str, optional): Specific (already parametrized) url to
                fetch the resource from

        Returns:
            :class:`.ApiResource`: Fetched resource

        Raises:
            Exception: If neither an ``_id`` nor a ``specific_url`` was provided
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        if specific_url is None:
            if _id is None:
                raise Exception('[{}] Provide an _id or a specific url for "from_id" method'.format(cls.resource))
            url = '/{}/{}'.format(cls.resource, _id)
        else:
            url = specific_url
        resp = client.request(url, method=requests.get,
                              message_prefix='From id {}'.format(cls.resource))

        resp_json = parse_json(resp)
        if _id is not None:
            logger.info('[Fetch {} OK] by id: "{}"'.format(cls.__name__, _id))
        else:
            logger.info('[Fetch {} OK] by url: "{}"'.format(cls.__name__, specific_url))

        return resp_json

    @classmethod
    def _list(cls, all: bool = True, project_id: str = None):
        """List all available instances of this resource type on the platform.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            dict: Fetched resources
        """
        if project_id:
            url = '/projects/{}/{}'.format(project_id, cls.resource)
        else:
            url = '/{}'.format(cls.resource)
        if all:
            return get_all_results(client, url, method=requests.get)
        else:
            resources = client.request(url, method=requests.get,
                                       message_prefix='List {}'.format(cls.resource))
            return parse_json(resources)['items']

    # def edit(self, **kwargs):
    #     """Edit a resource on the platform. You simply pass the function a
    #     dictionary of all the fields you want to update (as kwargs), with
    #     the name of the field as key and the new data for the field as value.

    #     .. note::

    #         The parameters you can update can be listed by calling:

    #         .. code-block:: python

    #             print(my_resource.resource_params)

    #     Returns:
    #         dict: Updated resource data

    #     Raises:
    #         PrevisionException: Any error while updating data on the platform
    #             or parsing result
    #     """
    #     update_fields = {
    #         'uid': self._id
    #     }

    #     for param in self.resource_params:
    #         if kwargs.get(param):
    #             update_fields[param] = kwargs[param]
    #     url = '/{}'.format(self.resource)
    #     resp = client.request(url,
    #                           body=update_fields,
    #                           method=requests.put)

    #     resp_json = parse_json(resp)

    #     logger.debug('[{}] {}'.format(self.resource, resp_json['message']))

    #     return resp_json['status'] == 200



# -*- coding: utf-8 -*-
from __future__ import print_function
from typing import Dict, Union

from pandas.core.frame import DataFrame
from previsionio.usecase_config import TypeProblem
import json
import uuid
import requests
import pandas as pd
from functools import lru_cache

from .logger import logger
from .dataset import Dataset
# from .deployed_model import DeployedModel
from .prevision_client import client
from .api_resource import ApiResource
from .utils import parse_json, PrevisionException, zip_to_pandas
from .prediction import ValidationPrediction


class Model(ApiResource):
    """ A Model object is generated by Prevision AutoML plateform when you launch a use case.
    All models generated by Prevision.io are deployable in our Store

    With this Model class, you can select the model with the optimal hyperparameters
    that responds to your buisiness requirements, then you can deploy it
    as a real-time/batch endpoint that can be used for a web Service.

    Args:
        _id (str): Unique id of the model
        usecase_version_id (str): Unique id of the usecase version of the model
        name (str, optional): Name of the model (default: ``None``)
    """

    def __init__(self, _id: str, usecase_version_id: str, project_id: str, model_name: str = None,
                 deployable: bool = False, **other_params):
        """ Instantiate a new :class:`.Model` object to manipulate a model resource on the platform. """
        super().__init__(_id=_id)
        self._id = _id
        self.usecase_version_id = usecase_version_id
        self.project_id = project_id
        self.name = model_name
        self.tags = {}
        self.deployable = deployable

        for k, v in other_params.items():
            self.__setattr__(k, v)

    def __repr__(self):
        return str(self._id)

    def __str__(self):
        """ Show up the Model object attributes.

        Returns:
            str: JSON-formatted info
        """
        args_to_show = {k: self.__dict__[k]
                        for k in self.__dict__
                        if all(map(lambda x: x not in k.lower(),
                                   ["event", "compositiondetails"])
                               )
                        }

        return json.dumps(args_to_show, sort_keys=True, indent=4, separators=(',', ': '))

    @classmethod
    def from_id(cls, _id: str) -> Union[
            'RegressionModel',
            'ClassificationModel',
            'MultiClassificationModel',
            'TextSimilarityModel']:
        """Get a usecase from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase to retrieve
            version (int, optional): Specific version of the usecase to retrieve
                (default: 1)

        Returns:
            :class:`.BaseUsecaseVersion`: Fetched usecase

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        end_point = '/models/{}'.format(_id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Model from id')
        model = json.loads(response.content.decode('utf-8'))
        training_type = TypeProblem(model.get('training_type', model.get('type_problem')))
        if training_type == TypeProblem.Regression:
            return RegressionModel(**model)
        elif training_type == TypeProblem.Classification:
            return ClassificationModel(**model)
        elif training_type == TypeProblem.MultiClassification:
            return MultiClassificationModel(**model)
        elif training_type == TypeProblem.TextSimilarity:
            return TextSimilarityModel(**model)
        else:
            raise PrevisionException('Training type {} not supported'.format(training_type))

    @property
    @lru_cache()
    def hyperparameters(self):
        """ Return the hyperparameters of a model.

        Returns:
            dict: Hyperparameters of the model
        """
        url = '/models/{}/hyperparameters/download'.format(self._id)
        response = client.request(
            endpoint=url,
            method=requests.get,
            message_prefix='Hyperparameters')
        return (json.loads(response.content.decode('utf-8')))

    def _predict_bulk(self,
                      dataset_id: str,
                      confidence: bool = False,
                      dataset_folder_id: str = None):
        """ (Util method) Private method used to handle bulk predict.

        .. note::

            This function should not be used directly. Use predict_from_* methods instead.

        Args:
            dataset_id (str): Unique id of the dataset to predict with
            confidence (bool, optional): Whether to predict with confidence estimator (default: ``False``)
            dataset_folder_id (str, optional): Unique id of the associated folder dataset to predict with,
                if need be (default: ``None``)

        Returns:
            str: A prediction object

        Raises:
            PrevisionException: Any error while starting the prediction on the platform or parsing the result
        """
        data = {
            'dataset_id': dataset_id,
            'model_id': self._id,
            'confidence': str(confidence).lower(),
        }

        if dataset_folder_id is not None:
            data['folder_dataset_id'] = dataset_folder_id

        predict_start = client.request('/usecase-versions/{}/predictions'.format(self.usecase_version_id),
                                       method=requests.post,
                                       data=data,
                                       message_prefix='Bulk predict')
        predict_start_parsed = parse_json(predict_start)

        if '_id' not in predict_start_parsed:
            err = 'Error starting prediction: {}'.format(predict_start_parsed)
            logger.error(err)
            raise PrevisionException(err)
        return ValidationPrediction(**predict_start_parsed)

    def predict_from_dataset(self, dataset: Dataset,
                             confidence: bool = False,
                             dataset_folder: Dataset = None) -> ValidationPrediction:
        """ Make a prediction for a dataset stored in the current active [client]
        workspace (using the current SDK dataset object).

        Args:
            dataset (:class:`.Dataset`): Dataset resource to make a prediction for
            confidence (bool, optional): Whether to predict with confidence values (default: ``False``)
            dataset_folder (:class:`.Dataset`, None): Matching folder dataset resource for the prediction,
                if necessary

        Returns:
            ``pd.DataFrame``: Prediction object
        """

        prediction = self._predict_bulk(dataset.id,
                                        confidence=confidence,
                                        dataset_folder_id=dataset_folder.id if dataset_folder else None)

        return prediction

    def predict(self, df: DataFrame, confidence: bool = False, prediction_dataset_name: str = None) -> pd.DataFrame:
        """ Make a prediction in a Scikit-learn blocking style.

        .. warning::

            For large dataframes and complex (blend) models, this can be slow (up to 1-2 hours). Prefer using
            this for simple models and small dataframes or use option ``use_best_single = True``.

        Args:
            df (``pd.DataFrame``): A ``pandas`` dataframe containing the testing data
            confidence (bool, optional): Whether to predict with confidence estimator (default: ``False``)

        Returns:
            ``pd.DataFrame``: Prediction results dataframe
        """
        if prediction_dataset_name is None:
            prediction_dataset_name = 'test_{}_{}'.format(self.name, str(uuid.uuid4())[-6:])

        dataset = Dataset._new(self.project_id, prediction_dataset_name, dataframe=df)

        prediction = self._predict_bulk(dataset.id,
                                        confidence=confidence)

        return prediction.get_result()

    def enable_deploy(self):
        data = {"deploy": True}
        url = '/models/{}'.format(self._id)
        response = client.request(url, requests.put, data=data,
                                  message_prefix="Enable deploy")
        self.deployable = True
        res = parse_json(response)
        return res

    def disable_deploy(self):
        data = {"deploy": False}
        url = '/models/{}'.format(self._id)
        response = client.request(url, requests.put, data=data,
                                  message_prefix="Disable deploy")
        self.deployable = False
        res = parse_json(response)
        return res

    # def deploy(self) -> DeployedModel:
    #     """ (Not Implemented yet) Deploy the model as a REST API app.
    #
    #     Keyword Arguments:
    #         app_type {enum} -- it can be 'model', 'notebook', 'shiny', 'dash' or 'node' application
    #
    #     Returns:
    #         str: Path of the deployed application
    #     """
    #     raise NotImplementedError


class ClassicModel(Model):

    @property
    @lru_cache()
    def feature_importance(self) -> pd.DataFrame:
        """ Return a dataframe of feature importances for the given model features, with their corresponding
        scores (sorted by descending feature importance scores).

        Returns:
            ``pd.DataFrame``: Dataframe of feature importances

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing the result
        """
        endpoint = '/models/{}/features-importances/download'.format(self._id)
        response = client.request(endpoint=endpoint,
                                  method=requests.get,
                                  message_prefix='Fetch feature importance')

        df_feat_importance = zip_to_pandas(response)

        return df_feat_importance.sort_values(by="importance", ascending=False)

    @property
    def cross_validation(self) -> pd.DataFrame:
        """ Get model's cross validation dataframe.

        Returns:
            ``pd.Dataframe``: Cross-validation dataframe
        """
        logger.debug('getting cv, model_id: {}'.format(self.id))
        url = '/models/{}/cross-validation/download'.format(self._id)
        cv_response = client.request(url,
                                     requests.get,
                                     message_prefix="Fetch cross validation")
        df_cv = zip_to_pandas(cv_response)

        return df_cv

    def chart(self):
        """ Return chart analysis information for a model.

        Returns:
            dict: Chart analysis results

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing the result
        """
        endpoint = '/models/{}/analysis'.format(self._id)
        response = client.request(
            endpoint=endpoint,
            method=requests.get,
            message_prefix='Model analysis')

        result = (json.loads(response.content.decode('utf-8')))

        # drop chart-related information
        return result

    def predict_single(self, data: Dict, confidence: bool = False, explain: bool = False):
        """ Make a prediction for a single instance. Use :py:func:`predict_from_dataset_name` or predict methods
        to predict multiple instances at the same time (it's faster).

        Args:
            data (dict): Features names and values (without target feature) - missing feature keys
                will be replaced by nans
            confidence (bool, optional): Whether to predict with confidence values (default: ``False``)
            explain (bool, optional): Whether to explain prediction (default: ``False``)


        .. note::

            You can set both ``confidence`` and ``explain`` to true.

        Returns:
            dict: Dictionary containing the prediction result

            .. note::

                The prediction format depends on the problem type (regression, classification, etc...)
        """
        payload = {
            'features': {
                str(k): v for k, v in data.items() if str(v) != 'nan'
            },
            'explain': explain,
            'confidence': confidence,
            'best': False,
            'model_id': self._id
        }

        logger.debug('[Predict Unit] sending payload ' + str(payload))
        url = '/usecase-versions/{}/unit-prediction'.format(self.usecase_version_id)
        response = client.request(url,
                                  requests.post,
                                  data=payload,
                                  content_type='application/json',
                                  message_prefix="Unit predict")
        response_json = parse_json(response)
        return response_json


class ClassificationModel(ClassicModel):
    """ A model object for a (binary) classification usecase, i.e. a usecase where the target
    is categorical with exactly 2 modalities.

    Args:
        _id (str): Unique id of the model
        uc_id (str): Unique id of the usecase of the model
        uc_version (str, int): Version of the usecase of the model (either an integer for a specific
            version, or "last")
        name (str, optional): Name of the model (default: ``None``)
    """

    def __init__(self, _id, usecase_version_id, **other_params):
        """ Instantiate a new :class:`.ClassificationModel` object to manipulate a classification model
        resource on the platform. """
        super().__init__(_id, usecase_version_id, **other_params)
        self._predict_threshold = 0.5

    @property
    @lru_cache()
    def optimal_threshold(self):
        """ Get the value of threshold probability that optimizes the F1 Score.

        Returns:
            float: Optimal value of the threshold (if it not a classification problem it returns ``None``)

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing the result
        """
        return self.chart()["optimal_proba"]

    def get_dynamic_performances(self, threshold: float = 0.5):
        """ Get model performance for the given threshold.

        Args:
            threshold (float, optional): Threshold to check the model's performance for (default: 0.5)

        Returns:
            dict: Model classification performance dict with the following keys:

                - ``confusion_matrix``
                - ``accuracy``
                - ``precision``
                - ``recall``
                - ``f1_score``

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing the result
        """
        threshold = float(threshold)
        if any((threshold > 1, threshold < 0)):
            err = 'threshold value has to be between 0 and 1'
            logger.error(err)
            raise ValueError(err)

        result = dict()
        query = '?threshold={}'.format(str(threshold))
        endpoint = '/models/{}/dynamic-analysis{}'.format(self._id, query)

        response = client.request(endpoint=endpoint,
                                  method=requests.get,
                                  message_prefix='Dynamic analysis')
        resp = json.loads(response.content.decode('utf-8'))

        result['confusion_matrix'] = resp["confusion_matrix"]
        for metric in ['accuracy', 'precision', 'recall', 'f1Score']:
            result[metric] = resp["score"][metric]

        return result


class RegressionModel(ClassicModel):
    """ A model object for a regression usecase, i.e. a usecase where the target is numerical.

    Args:
        _id (str): Unique id of the model
        uc_id (str): Unique id of the usecase of the model
        uc_version (str, int): Version of the usecase of the model (either an integer for a specific
            version, or "last")
        name (str, optional): Name of the model (default: ``None``)
    """


class MultiClassificationModel(ClassicModel):
    """ A model object for a multi-classification usecase, i.e. a usecase where the target
    is categorical with strictly more than 2 modalities.

    Args:
        _id (str): Unique id of the model
        uc_id (str): Unique id of the usecase of the model
        uc_version (str, int): Version of the usecase of the model (either an integer for a specific
            version, or "last")
        name (str, optional): Name of the model (default: ``None``)
    """


class TextSimilarityModel(Model):

    def _predict_bulk(self,
                      queries_dataset_id: str,
                      queries_dataset_content_column: str,
                      top_k: int,
                      matching_id_description_column: str = None):
        """ (Util method) Private method used to handle bulk predict.

        .. note::

            This function should not be used directly. Use predict_from_* methods instead.

        Args:
            queries_dataset_id (str): Unique id of the quries dataset to predict with
            queries_dataset_content_column (str): Content queries column name
            queries_dataset_matching_id_description_column (str): Matching id description column name
            top_k (integer): Number of the nearest description to predict
        Returns:
            str: A prediction job ID

        Raises:
            PrevisionException: Any error while starting the prediction on the platform or parsing the result
        """
        data = {
            'model_id': self._id,
            'queries_dataset_id': queries_dataset_id,
            'queries_dataset_content_column': queries_dataset_content_column,
            'top_k': top_k
        }
        if matching_id_description_column:
            data['queries_dataset_matching_id_description_column'] = matching_id_description_column
        endpoint = '/usecase-versions/{}/predictions'.format(self.usecase_version_id)
        predict_start = client.request(endpoint,
                                       method=requests.post,
                                       data=data,
                                       message_prefix='Text-similarity bulk predict')
        predict_start_parsed = parse_json(predict_start)

        if '_id' not in predict_start_parsed:
            err = 'Error starting prediction: {}'.format(predict_start_parsed)
            logger.error(err)
            raise PrevisionException(err)

        return ValidationPrediction(**predict_start_parsed)

    def predict_from_dataset(self, queries_dataset: Dataset, queries_dataset_content_column: str, top_k: int = 10,
                             queries_dataset_matching_id_description_column: str = None) -> ValidationPrediction:
        """ Make a prediction for a dataset stored in the current active [client]
        workspace (using the current SDK dataset object).

        Args:
            dataset (:class:`.Dataset`): Dataset resource to make a prediction for
            queries_dataset_content_column (str): Content queries column name
            top_k (integer): Number of the nearest description to predict
            queries_dataset_matching_id_description_column (str): Matching id description column name

        Returns:
            ``pd.DataFrame``: Prediction object
        """
        prediction = self._predict_bulk(queries_dataset.id,
                                        queries_dataset_content_column,
                                        top_k=top_k,
                                        matching_id_description_column=queries_dataset_matching_id_description_column)
        return prediction

    def predict(self, df: DataFrame, queries_dataset_content_column: str, top_k: int = 10,
                queries_dataset_matching_id_description_column: str = None,
                prediction_dataset_name: str = None) -> Union[pd.DataFrame, None]:
        """ Make a prediction for a dataset stored in the current active [client]
        workspace (using the current SDK dataset object).

        Args:
            dataset (:class:`.Dataset`): Dataset resource to make a prediction for
            queries_dataset_content_column (str): Content queries column name
            top_k (integer): Number of the nearest description to predict
            queries_dataset_matching_id_description_column (str): Matching id description column name

        Returns:
            ``pd.DataFrame``: Prediction results dataframe
        """
        if prediction_dataset_name is None:
            prediction_dataset_name = 'test_{}_{}'.format(self.name, str(uuid.uuid4())[-6:])

        dataset = Dataset._new(self.project_id, prediction_dataset_name, dataframe=df)
        prediction = self._predict_bulk(dataset.id,
                                        queries_dataset_content_column,
                                        top_k=top_k,
                                        matching_id_description_column=queries_dataset_matching_id_description_column)

        return prediction.get_result()



from enum import Enum
import operator
from typing import Dict, List, Tuple, Union
import requests
import uuid
import json
import zipfile
import pandas as pd
import os
import numpy as np
from requests.models import Response
from . import logger, config
import datetime
from math import ceil


def zip_to_pandas(pred_response: requests.Response, separator=None) -> pd.DataFrame:
    temp_zip_path = '/tmp/ziptmp{}.zip'.format(str(uuid.uuid4()))

    with open(temp_zip_path, 'wb') as temp:
        for chunk in pred_response.iter_content(chunk_size=1024):
            if chunk:
                temp.write(chunk)

        temp.seek(0)

    with zipfile.ZipFile(temp_zip_path, 'r') as pred_zip:
        names = pred_zip.namelist()
        pred_zip.extractall('/tmp')
        pred_csv_path = '/tmp/' + names[0]
        data = pd.read_csv(pred_csv_path, sep=separator, engine='python')
        os.remove(pred_csv_path)

    return data


class PrevisionException(Exception):
    pass


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.datetime64):
            return (datetime.datetime.fromtimestamp(obj.astype(datetime.datetime) // 10 ** 9)
                    .strftime('%Y-%m-%dT%H:%M:%s'))
        elif isinstance(obj, pd.Timestamp):
            return obj.strftime('%Y-%m-%dT%H:%M:%s')
        else:
            return super(NpEncoder, self).default(obj)


def parse_json(json_response: Response) -> Dict:
    try:
        return json_response.json()
    except Exception as e:
        logger.error('[JSON PARSE] ' + json_response.text + '  --  ' + e.__repr__())
        raise


def to_json(obj):
    if isinstance(obj, Enum):
        return to_json(obj.value)
    elif isinstance(obj, list):
        obj_list = []
        for e in obj:
            obj_list.append(to_json(e))
        return obj_list
    elif isinstance(obj, dict):
        obj_d = {}
        for key, value in obj.items():
            obj_d[key] = to_json(value)
        return obj_d
    elif hasattr(obj, '__dict__'):
        obj_dict = {}
        for key, value in obj.__dict__.items():
            if value:
                if hasattr(obj, 'config') and key in obj.config:
                    key = obj.config[key]
                obj_dict[key] = to_json(value)
        return obj_dict
    return obj


def get_pred_from_multiclassification(row, pred_prefix: str = 'pred_'):
    d = row.to_dict()
    preds_probas = {k: float(v) for k, v in d.items() if pred_prefix in k}
    pred = max(preds_probas.items(), key=operator.itemgetter(1))[0]
    pred = pred.replace(pred_prefix, '')
    return pred


class EventTuple():
    def __init__(self, name: str,
                 success_checks: Union[Tuple[str, str], List[Tuple[str, str]]] = ('state', 'done'),
                 fail_checks: Union[Tuple[str, str], List[Tuple[str, str]]] = ('state', 'failed')):
        assert name is not None
        self.name = name

        assert success_checks is not None
        self.success_checks = success_checks
        if isinstance(success_checks, Tuple):
            self.success_checks = [success_checks]
        assert len(set([key for key, _ in self.success_checks])) == len(
            self.success_checks), "requiring same key for success check"

        assert fail_checks is not None
        self.fail_checks = fail_checks
        if isinstance(fail_checks, Tuple):
            self.fail_checks = [fail_checks]

    def is_failure(self, json: Dict):
        """
            if any fail condition pass, it's a failure
        """
        for key, val in self.fail_checks:
            if json.get(key) == val:
                return True
        return False

    def is_success(self, json: Dict):
        """
            all success condition need to be there to pass
        """
        result = True
        for key, val in self.success_checks:
            if json.get(key) != val:
                return False
        return result


def is_null_value(value) -> bool:
    is_null = str(value) == 'null'
    is_nan = pd.isnull(value)
    return is_nan or is_null


def get_all_results(client, endpoint: str, method) -> List[Dict]:
    resources = []
    batch: requests.Response = client.request(endpoint, method=method)
    json = parse_json(batch)
    meta = json['metaData']
    total_items = meta['totalItems']
    rows_per_page = meta['rowsPerPage']
    n_pages = ceil(total_items / rows_per_page)
    for n in range(1, n_pages + 1):
        url = endpoint + "?page={}".format(n)
        batch = client.request(url, method=method)
        resources.extend(parse_json(batch)['items'])
    return resources


def handle_error_response(
    resp: Response,
    url: str,
    data: Union[Dict, List] = None,
    files: Dict = None,
    message_prefix: str = None,
    n_tries: int = 1,
    additional_log: str = None,
):
    if resp.status_code not in config.success_codes:
        message = "Error {}: '{}' reaching url: '{}'".format(
            resp.status_code, resp.text, url)
        if n_tries > 1:
            message += " after {} tries".format(n_tries)
        if data:
            message += " with data: {}".format(data)
        if files:
            message += " with files: {}".format(files)
        if message_prefix:
            message = message_prefix + ' failure\n' + message
        logger.error(message)
        if additional_log:
            logger.error(additional_log)
        raise PrevisionException(message)



from typing import List
import time
import requests

from . import config
from .logger import logger
from .api_resource import ApiResource
from . import client
from .usecase_version import BaseUsecaseVersion
from .utils import parse_json, PrevisionException, get_all_results
from .prediction import DeploymentPrediction
from .dataset import Dataset


class UsecaseDeployment(ApiResource):
    """ UsecaseDeployment objects represent usecase deployment resource that will be explored by Prevision.io platform.
    """

    resource = 'model-deployments'

    def __init__(self, _id: str, name: str, usecase_id, current_version,
                 versions, deploy_state, access_type, project_id, training_type, models, url=None,
                 **kwargs):

        self.name = name
        self._id = _id
        #self.usecase_version_id = usecase_version_id
        self.usecase_id = usecase_id
        self.current_version = current_version
        self.versions = versions
        self._deploy_state = deploy_state
        self.access_type = access_type
        self.project_id = project_id
        self.training_type = training_type
        self.models = models
        self.url = url

        self._run_state = kwargs.pop("main_model_run_state", kwargs.pop("run_state", "error"))
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    @classmethod
    def from_id(cls, _id: str):
        """Get a deployed usecase from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase version to retrieve

        Returns:
            :class:`.UsecaseDeployment`: Fetched deployed usecase

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        url = '/{}/{}'.format('deployments', _id)
        result = super()._from_id(_id=_id, specific_url=url)
        logger.debug(result)
        return cls(**result)

    @property
    def deploy_state(self):
        usecase_deployment = self.from_id(self._id)
        return usecase_deployment._deploy_state

    @property
    def run_state(self):
        usecase_deployment = self.from_id(self._id)
        return usecase_deployment._run_state

    @classmethod
    def list(cls, project_id: str, all: bool = True) -> List['UsecaseDeployment']:
        """ List all the available usecase in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.UsecaseDeployment` objects rather
            than plain dictionaries with the corresponding data.

        Args:
            project_id (str): project id
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.UsecaseDeployment`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [UsecaseDeployment(**usecase_deployment) for usecase_deployment in resources]

    @classmethod
    def _new(cls, project_id: str, name: str, main_model, challenger_model=None, access_type: str = 'public'):
        """ Create a new usecase deployement object on the platform.

        Args:
            project_id (str): project id
            name (str): usecase deployement name
            main_model: main model
            challenger_model (optional): challenger model. main and challenger models should be in the same usecase
            access_type (str, optional): public/ fine_grained/ private

        Returns:
            :class:`.UsecaseDeployment`: The registered usecase deployment object in the current project

        Raises:
            PrevisionException: Any error while creating usecase deployment to the platform
                or parsing the result
            Exception: For any other unknown error
        """

        if access_type not in ['public', 'fine_grained', 'private']:
            raise PrevisionException('access type must be public, fine_grained or private')
        main_model_usecase_version_id = main_model.usecase_version_id
        main_usecase = BaseUsecaseVersion._from_id(main_model_usecase_version_id)
        main_usecase_id = main_usecase['usecase_id']
        data = {
            'name': name,
            'usecase_id': main_usecase_id,
            'main_model_usecase_version_id': main_model_usecase_version_id,
            'main_model_id': main_model._id,
            'access_type': access_type
        }

        if challenger_model:
            challenger_model_usecase_version_id = challenger_model.usecase_version_id
            challenger_usecase = BaseUsecaseVersion._from_id(main_model_usecase_version_id)
            challenger_usecase_id = challenger_usecase['usecase_id']
            if main_usecase_id != challenger_usecase_id:
                raise PrevisionException('main and challenger models must be from the same usecase')
            data['challenger_model_usecase_version_id'] = challenger_model_usecase_version_id
            data['challenger_model_id'] = challenger_model._id

        url = '/projects/{}/model-deployments'.format(project_id)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='UsecaseDeployment creation')
        json_resp = parse_json(resp)
        usecase_deployment = cls.from_id(json_resp['_id'])
        return usecase_deployment

    def new_version(self, name: str, main_model, challenger_model=None):
        """ Create a new usecase deployement version.

        Args:
            name (str): usecase deployement name
            main_model: main model
            challenger_model (optional): challenger model. main and challenger models should be in the same usecase

        Returns:
            :class:`.UsecaseDeployment`: The registered usecase deployment object in the current project

        Raises:
            PrevisionException: Any error while creating usecase deployment to the platform
                or parsing the result
            Exception: For any other unknown error
        """
        main_model_usecase_version_id = main_model.usecase_version_id
        main_usecase = BaseUsecaseVersion._from_id(main_model_usecase_version_id)
        main_usecase_id = main_usecase['usecase_id']
        data = {
            'name': name,
            'main_model_usecase_version_id': main_model_usecase_version_id,
            'main_model_id': main_model._id,
        }
        if challenger_model:
            challenger_model_usecase_version_id = challenger_model.usecase_version_id
            challenger_usecase = BaseUsecaseVersion._from_id(main_model_usecase_version_id)
            challenger_usecase_id = challenger_usecase['usecase_id']
            if main_usecase_id != challenger_usecase_id:
                raise PrevisionException('main and challenger models must be from the same usecase')
            data['challenger_model_usecase_version_id'] = challenger_model_usecase_version_id
            data['challenger_model_id'] = challenger_model._id

        url = '/deployments/{}/versions'.format(self._id)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='UsecaseDeployment creation new version')
        json_resp = parse_json(resp)
        usecase_deployment = self.from_id(json_resp['_id'])
        return usecase_deployment

    def delete(self):
        """Delete a usecase deployement from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        resp = client.request(endpoint='/deployments/{}'.format(self.id),
                              method=requests.delete,
                              message_prefix='UsecaseDeployment delete')
        return resp

    def wait_until(self, condition, timeout: float = config.default_timeout):
        """ Wait until condition is fulfilled, then break.

        Args:
            condition (func: (:class:`.BaseUsecaseVersion`) -> bool.): Function to use to check the
                break condition
            raise_on_error (bool, optional): If true then the function will stop on error,
                otherwise it will continue waiting (default: ``True``)
            timeout (float, optional): Maximal amount of time to wait before forcing exit

        Example::

            usecase.wait_until(lambda usecasev: len(usecasev.models) > 3)

        Raises:
            PrevisionException: If the resource could not be fetched or there was a timeout.
        """
        t0 = time.time()
        while True:
            if timeout is not None and time.time() - t0 > timeout:
                raise PrevisionException('timeout while waiting on {}'.format(condition))
            try:
                if condition(self):
                    break
                elif self.deploy_state == 'failed' or self.run_state == 'failed':
                    raise PrevisionException('Resource failed while waiting')
            except PrevisionException as e:
                logger.warning(e.__repr__())
                raise

            time.sleep(config.scheduler_refresh_rate)

    def create_api_key(self):
        """Create an api key of the usecase deployement from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        print('/deployments/{}/api-keys'.format(self.id))
        resp = client.request(endpoint='/deployments/{}/api-keys'.format(self.id),
                              method=requests.post,
                              message_prefix='UsecaseDeployment create api key')
        resp = parse_json(resp)
        return resp

    def get_api_keys(self):
        """Fetch the api keys of the usecase deployement from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        print('/deployments/{}/api-keys'.format(self.id))
        resp = client.request(endpoint='/deployments/{}/api-keys'.format(self.id),
                              method=requests.get,
                              message_prefix='UsecaseDeployment get api key')
        resp = parse_json(resp)
        return resp

    def predict_from_dataset(self, dataset: Dataset) -> DeploymentPrediction:
        """ Make a prediction for a dataset stored in the current active [client]
        workspace (using the current SDK dataset object).

        Args:
            dataset (:class:`.Dataset`): Dataset resource to make a prediction for

        Returns:
            ``pd.DataFrame``: Prediction object
        """
        if self.training_type not in ['regression', 'classification', 'multiclassification']:
            PrevisionException('Prediction not supported yet for training type {}', self.training_type)
        data = {
            'dataset_id': dataset._id,
        }

        predict_start = client.request('/deployments/{}/deployment-predictions'.format(self._id),
                                       method=requests.post,
                                       data=data,
                                       message_prefix='Bulk predict')
        predict_start_parsed = parse_json(predict_start)
        return DeploymentPrediction(**predict_start_parsed)

    def list_predictions(self) -> List[DeploymentPrediction]:
        """ List all the available predictions in the current active [client] workspace.

        Returns:
            list(:class:`.DeploymentPrediction`): Fetched deployed predictions objects
        """
        end_point = '/deployments/{}/deployment-predictions'.format(self._id)
        predictions = get_all_results(client, end_point, method=requests.get)
        return [DeploymentPrediction(**prediction) for prediction in predictions]



# -*- coding: utf-8 -*-
from __future__ import print_function

from previsionio.text_similarity import TextSimilarity
from previsionio.supervised import Supervised
from previsionio.timeseries import TimeSeries
from previsionio.usecase_config import DataType, TypeProblem
from typing import Dict, List, Type, Union
import requests

from .prevision_client import client
from .utils import parse_json
from .api_resource import ApiResource


def get_usecase_version_class(training_type: TypeProblem, data_type: DataType) -> Union[Type[TextSimilarity], Type[Supervised], Type[TimeSeries]]:
    """ Get the type of UsecaseVersion class used by this Usecase

    Returns:
        (:type:`.TextSimilarity` | :type:`.Supervised` | :type:`.TimeSeries`): Type of UsecaseVersion
    """
    default: Dict[DataType, Union[Type[Supervised], Type[TimeSeries]]] = {
        DataType.Tabular: Supervised,
        DataType.TimeSeries: TimeSeries
    }
    class_dict = {
        TypeProblem.TextSimilarity: {DataType.Tabular: TextSimilarity},
    }
    class_type = class_dict.get(training_type, default).get(data_type)
    assert class_type is not None
    return class_type


class Usecase(ApiResource):
    """ A Usecase

    Args:
        _id (str): Unique id of the usecase
        name (str): Name of the usecase

    """

    resource = 'usecases'

    def __init__(self, **usecase_info):
        super().__init__(**usecase_info)
        self._id: str = usecase_info['_id']
        self.name: str = usecase_info['name']
        self.project_id: str = usecase_info['project_id']
        self.training_type: TypeProblem = TypeProblem(usecase_info['training_type'])
        self.data_type: DataType = DataType(usecase_info['data_type'])

    @classmethod
    def from_id(cls, _id: str) -> 'Usecase':
        """Get a usecase from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase version to retrieve

        Returns:
            :class:`.Usecase`: Fetched usecase

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return cls(**super()._from_id(specific_url='/{}/{}'.format(cls.resource, _id)))

    @classmethod
    def list(cls, project_id: str, all: bool = True) -> List['Usecase']:
        """ List all the available usecase in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Usecase` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            project_id (str): project id
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Usecase`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**conn_data) for conn_data in resources]

    @property
    def usecase_version_class(self) -> Union[Type[TextSimilarity], Type[Supervised], Type[TimeSeries]]:
        """ Get the type of UsecaseVersion class used by this Usecase

        Returns:
            (:type:`.TextSimilarity` | :type:`.Supervised` | :type:`.TimeSeries`): Type of UsecaseVersion
        """
        return get_usecase_version_class(self.training_type, self.data_type)

    @property
    def latest_version(self) -> Union[TextSimilarity, Supervised, TimeSeries]:
        """Get the latest version of this use case.

        Returns:
            (:class:`.TextSimilarity` | :class:`.Supervised` | :class:`.TimeSeries`):
            latest UsecaseVersion in this Usecase
        """
        end_point = '/{}/{}/versions'.format(self.resource, self._id)
        format = {
            "rowsPerPage": 1,
            "sortBy": "created_at",
            "descending": True
        }

        response = client.request(endpoint=end_point,
                                  format=format,
                                  method=requests.get,
                                  message_prefix='Latest usecase version')
        res = parse_json(response)
        assert len(res['items']) == 1
        return self.usecase_version_class(**res['items'][0])

    @property
    def versions(self) -> List[Union[TextSimilarity, Supervised, TimeSeries]]:
        """Get the list of all versions for the current use case.

        Returns:
            list(:class:`.TextSimilarity` | :class:`.Supervised` | :class:`.TimeSeries`):
            List of the usecase versions (as JSON metadata)
        """
        end_point = '/{}/{}/versions'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Usecase versions listing')
        res = parse_json(response)
        return [self.usecase_version_class(**val) for val in res['items']]

    def delete(self):
        """ Delete a usecase from the actual [client] workspace."""
        _ = client.request(endpoint='/usecases/{}'.format(self._id),
                           method=requests.delete,
                           message_prefix='Usecase deletion')



# -*- coding: utf-8 -*-
from __future__ import print_function
from typing import Dict, Union

import requests
from previsionio.utils import EventTuple, parse_json, to_json
from . import TrainingConfig
from .usecase_config import DataType, UsecaseConfig, ColumnConfig, TypeProblem
from .usecase_version import ClassicUsecaseVersion
from .metrics import Regression
from .model import RegressionModel
from .dataset import Dataset
from .prevision_client import client
import previsionio as pio


class TimeWindowException(Exception):
    pass


class TimeWindow(UsecaseConfig):
    """
    A time window object for representing either feature derivation window periods or
    forecast window periods.

    Args:
        derivation_start (int): Start of the derivation window (must be < 0)
        derivation_end (int): End of the derivation window (must be < 0)
        forecast_start (int): Start of the forecast window (must be > 0)
        forecast_end (int): End of the forecast window (must be > 0)
    """
    config = {
        'derivation_start': 'start_dw',
        'derivation_end': 'end_dw',
        'forecast_start': 'start_fw',
        'forecast_end': 'end_fw',
    }

    def __init__(self, derivation_start: int, derivation_end: int, forecast_start: int, forecast_end: int):
        """Instantiate a new :class:`.TimeWindow` util object."""

        if not derivation_start < derivation_end or not forecast_start < forecast_end:
            raise TimeWindowException('start must be smaller than end')

        if derivation_start > 0 or derivation_end > 0:
            raise TimeWindowException('derivation window bounds must be negative')

        if forecast_start < 0 or forecast_end < 0:
            raise TimeWindowException('forecast window bounds must be positive')

        if not(derivation_end <= ((2 * forecast_start) - forecast_end)):
            raise TimeWindowException('derivation_end must be smaller than (2 * forecast_start) - forecast_end')
        # Not valid anymore :
        # if not abs(derivation_end) >= (forecast_end - forecast_start):
        #     raise TimeWindowException('end of derivation window must be smaller than end of forecast window')

        self.derivation_start = derivation_start
        self.forecast_start = forecast_start
        self.derivation_end = derivation_end
        self.forecast_end = forecast_end


class TimeSeries(ClassicUsecaseVersion):
    """ A supervised usecase version, for timeseries data """

    training_type = TypeProblem.Regression
    metric_type = Regression
    default_metric = Regression.RMSE
    data_type = DataType.TimeSeries
    model_class = RegressionModel

    def __init__(self, **usecase_info):
        self.holdout_dataset_id: Union[str, None] = usecase_info.get('holdout_dataset_id', None)
        self.time_window = TimeWindow.from_dict(usecase_info['usecase_version_params']['timeseries_values'])

        super().__init__(**usecase_info)

    @classmethod
    def from_id(cls, _id: str) -> 'TimeSeries':
        return cls(**super()._from_id(_id))

    @classmethod
    def load(cls, pio_file: str) -> 'TimeSeries':
        return cls(**super()._load(pio_file))

    @classmethod
    def _fit(cls, project_id: str,
             name: str,
             dataset: Dataset,
             column_config: ColumnConfig,
             time_window: TimeWindow,
             metric: Regression = None,
             holdout_dataset: Dataset = None,
             training_config: TrainingConfig = TrainingConfig()) -> 'TimeSeries':
        training_args = to_json(training_config)
        assert isinstance(training_args, Dict)

        training_args.update(to_json(column_config))
        training_args.update(to_json(time_window))

        if holdout_dataset:
            training_args['holdout_dataset_id'] = holdout_dataset.id

        if not metric:
            metric = cls.default_metric

        start_response = cls._start_usecase(project_id=project_id,
                                            name=name,
                                            dataset_id=dataset.id,
                                            data_type=cls.data_type,
                                            training_type=cls.training_type,
                                            metric=metric.value,
                                            **training_args)

        usecase = cls.from_id(start_response['_id'])
        events_url = '/{}/{}'.format(cls.resource, start_response['_id'])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(usecase.resource_id,
                                                cls.resource,
                                                EventTuple('USECASE_VERSION_UPDATE', ('state', 'running')),
                                                specific_url=events_url)
        return usecase

    def new_version(
        self,
        description: str = None,
        dataset: Dataset = None,
        column_config: ColumnConfig = None,
        time_window: TimeWindow = None,
        metric: Regression = None,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = TrainingConfig(),
    ):
        """ Start a time series usecase training to create a new version of the usecase (on the
        platform): the training configs are copied from the current version and then overridden
        for the given parameters.

        Args:
            description (str, optional): additional description of the version
            dataset (:class:`.Dataset`, :class:`.DatasetImages`, optional): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`, optional): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            time_window (:class: `.TimeWindow`, optional): a time window object for representing either feature
                derivation window periods or forecast window periods
            metric (metrics.Regression, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
        Returns:
            :class:`.TimeSeries`: Newly created text similarity usecase version object (new version)
        """

        if not dataset:
            dataset_id = self.dataset_id
        else:
            dataset_id = dataset.id

        if not column_config:
            column_config = self.column_config

        if not time_window:
            time_window = self.time_window

        if not metric:
            metric = Regression(self.metric)

        if not holdout_dataset:
            holdout_dataset_id = self.holdout_dataset_id
        else:
            holdout_dataset_id = holdout_dataset.id

        if not training_config:
            training_config = self.training_config

        training_args = to_json(training_config)
        assert isinstance(training_args, Dict)
        training_args.update(to_json(column_config))
        training_args.update(to_json(time_window))

        params = {
            'dataset_id': dataset_id,
            'metric': metric.value,
            'holdout_dataset': holdout_dataset_id,
            'training_type': self.training_type.value,
            'usecase_id': self._id,
            'parent_version': self.version,
            # 'nextVersion': max([v['version'] for v in self.versions]) + 1  FA: wait what ?
        }

        if description:
            params["description"] = description

        params.update(training_args)

        endpoint = "/usecases/{}/versions".format(self.usecase_id)

        resp = client.request(endpoint=endpoint,
                              data=params,
                              method=requests.post,
                              content_type='application/json',
                              message_prefix='Time series usecase start')
        json = parse_json(resp)

        usecase = self.from_id(json["_id"])

        events_url = '/{}/{}'.format(self.resource, json['_id'])
        assert client.event_manager is not None
        client.event_manager.wait_for_event(usecase.resource_id,
                                            self.resource,
                                            EventTuple('USECASE_VERSION_UPDATE', ('state', 'running')),
                                            specific_url=events_url)

        return usecase



# -*- coding: utf-8 -*-
from __future__ import print_function
from itertools import combinations
from io import BytesIO, StringIO
from previsionio.usecase_config import UsecaseState
import numpy as np
import pandas as pd
import os
import tempfile
import zipfile
import previsionio.utils
from . import client
from .utils import parse_json, zip_to_pandas, PrevisionException
from .api_resource import ApiResource
from . import logger
import previsionio as pio
import requests
from .datasource import DataSource

from typing import Dict, Union
from pandas import DataFrame, Series
FrameOrSeriesUnion = Union["DataFrame", "Series"]


class Dataset(ApiResource):

    """ Dataset objects represent data resources that will be explored by Prevision.io platform.

        In order to launch an auto ml process (see :class:`.BaseUsecase` class), we need to have
        the matching dataset stored in the related workspace.

        Within the platform they are stored in tabular form and are derived:

        - from files (CSV, ZIP)
        - or from a Data Source at a given time (snapshot) """

    resource = 'datasets'

    def __init__(self, _id: str, name: str, datasource: DataSource = None, _data: DataFrame = None,
                 describe_state: Dict = None, drift_state=None, embeddings_state=None, separator=',', **kwargs):

        super().__init__(_id=_id, datasource=datasource)
        self.name = name
        self._id = _id
        self.separator = separator
        self.datasource = datasource
        self.describe_state = describe_state
        self.drift_state = drift_state
        self.embeddings_state = embeddings_state
        self.other_params = kwargs
        self._data = _data

    def to_pandas(self) -> pd.DataFrame:
        """Load in memory the data content of the current dataset into a pandas DataFrame.

        Returns:
            ``pd.DataFrame``: Dataframe for the data object

        Raises:
            PrevisionException: Any error while fetching or parsing the data
        """
        if self._data is not None:
            return self._data
        else:
            response = client.request(endpoint='/{}/{}/download'.format(self.resource, self.id),
                                      method=requests.get,
                                      message_prefix='Dataset download')

            self._data = zip_to_pandas(response, separator=self.separator)

            return self._data

    data = property(to_pandas)

    @classmethod
    def from_id(cls, _id: str) -> 'Dataset':
        return cls(**super()._from_id(_id=_id))

    @classmethod
    def list(cls, project_id: str, all: bool = True):
        """ List all the available datasets in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Dataset` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Dataset`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**conn_data) for conn_data in resources]

    def update_status(self):
        url = '/{}/{}'.format(self.resource, self._id)
        dset_resp = client.request(url, method=requests.get, message_prefix='Dataset status update')
        dset_json = parse_json(dset_resp)
        self.describe_state = dset_json['describe_state']
        self.drift_state = dset_json['drift_state']
        self.embeddings_state = dset_json['embeddings_state']

    def get_describe_status(self):
        return UsecaseState(self.describe_state)

    def get_drift_status(self):
        return UsecaseState(self.drift_state)

    def get_embedding_status(self):
        return UsecaseState(self.embeddings_state)

    def delete(self):
        """Delete a dataset from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        resp = client.request(endpoint='/{}/{}'.format(self.resource, self.id),
                              method=requests.delete,
                              message_prefix='Dataset delete')
        return resp

    def start_embedding(self):
        """Starts the embeddings analysis of the dataset from the actual [client] workspace

        Raises:
            PrevisionException: DatasetNotFoundError
            requests.exceptions.ConnectionError: request error
        """

        resp = client.request(endpoint='/{}/{}/analysis'.format(self.resource, self.id),
                              method=requests.post,
                              message_prefix='Dataset start embedding')

        logger.info('Started embedding for dataset {}'.format(self.name))
        resp = parse_json(resp)
        return resp

    def get_embedding(self) -> Dict:
        """Gets the embeddings analysis of the dataset from the actual [client] workspace

        Raises:
            PrevisionException: DatasetNotFoundError
            requests.exceptions.ConnectionError: request error
        """
        resp = client.request(endpoint='/{}/{}/explorer'.format(self.resource, self.id),
                              method=requests.get,
                              message_prefix='Dataset get embedding')

        tensors_shape = parse_json(resp)['embeddings'][0]['tensorShape']
        labels_resp = client.request(endpoint='/{}/{}/explorer/labels.bytes'.format(self.resource, self.id),
                                     method=requests.get,
                                     message_prefix='Dataset labels bytes')
        tensors_resp = client.request(endpoint='/{}/{}/explorer/tensors.bytes'.format(self.resource, self.id),
                                      method=requests.get,
                                      message_prefix='Dataset tensors bytes')

        labels = pd.read_csv(StringIO(labels_resp.text), sep="\t")
        tensors = np.frombuffer(BytesIO(tensors_resp.content).read(),
                                dtype="float32").reshape(*tensors_shape)
        return {'labels': labels, 'tensors': tensors}

    def download(self, download_path: str = None):
        """Download the dataset from the platform locally.

        Args:
            download_path (str, optional): Target local directory path
                (if none is provided, the current working directory is
                used)

        Returns:
            str: Path the data was downloaded to

        Raises:
            PrevisionException: If dataset does not exist or if there
                was another error fetching or parsing data
        """
        endpoint = '/{}/{}/download'.format(self.resource, self.id)
        resp = client.request(endpoint=endpoint,
                              method=requests.get,
                              message_prefix='Dataset download')

        if resp._content is None:
            raise PrevisionException('could not download dataset')

        if not download_path:
            download_path = os.getcwd()
        path = os.path.join(download_path, self.name + ".zip")

        with open(path, "wb") as file:
            file.write(resp._content)

        return path

    @classmethod
    def _new(cls, project_id: str, name: str, datasource: DataSource = None,
             file_name: str = None, dataframe: DataFrame = None, **kwargs):
        """ Register a new dataset in the workspace for further processing.
        You need to provide either a datasource, a file name or a dataframe
        (only one can be specified).

        .. note::

            To start a new use case on a dataset, it has to be already
            registred in your workspace.

        Args:
            project_id (str): project id
            name (str): Registration name for the dataset
            datasource (:class:`.DataSource`, optional): A DataSource object used
                to import a remote dataset (if you want to import a specific dataset
                from an existent database, you need a datasource connector
                (:class:`.Connector` object) designed to point to the related data source)
            file_name (str, optional): Path to a file to upload as dataset
            dataframe (pd.DataFrame, optional): A ``pandas`` dataframe containing the
                data to upload

        Raises:
            Exception: If more than one of the keyword arguments ``datasource``, ``file_name``,
                ``dataframe`` was specified
            PrevisionException: Error while creating the dataset on the platform

        Returns:
            :class:`.Dataset`: The registered dataset object in the current workspace.
        """
        if any(a is not None and b is not None for a, b in combinations([datasource, file_name, dataframe], 2)):
            raise Exception('only one of [datasource, file_handle, data] must be specified')

        if not any([datasource is not None, file_name is not None, dataframe is not None]):
            raise Exception('at least one of [datasource, file_handle, data] must be specified')

        data = {
            'name': name,
        }

        if 'origin' in kwargs:
            valid_origin = ["pipeline_output", "pipeline_intermediate_file"]
            origin = kwargs.get('origin')
            if not isinstance(origin, str) or origin not in valid_origin:
                raise RuntimeError(f"invalid origin: {origin}")
            data['origin'] = origin

        files = {

        }
        request_url = '/projects/{}/{}/file'.format(project_id, cls.resource)
        create_resp = None
        if datasource is not None:
            request_url = '/projects/{}/{}/data-source'.format(project_id, cls.resource)
            data['datasource_id'] = datasource.id
            create_resp = client.request(request_url,
                                         data=data,
                                         method=requests.post,
                                         message_prefix='Dataset upload from datasource')

        elif dataframe is not None:
            file_ext = '.zip'

            with tempfile.NamedTemporaryFile(prefix=name, suffix='.csv') as temp:
                dataframe.to_csv(temp.name, index=False)

                file_name = temp.name.replace('.csv', file_ext)
                with zipfile.ZipFile(file_name, 'w') as zip_file:
                    zip_file.write(temp.name, arcname=name + '.csv')
                assert zip_file.filename is not None

                with open(zip_file.filename, 'rb') as f:
                    files['file'] = (os.path.basename(file_name), f, 'application/zip')
                    for k, v in data.items():
                        files[k] = (None, v)

                    create_resp = client.request(request_url,
                                                 data=data,
                                                 files=files,
                                                 method=requests.post,
                                                 message_prefix='Dataset upload from dataframe')

        elif file_name is not None:

            if zipfile.is_zipfile(file_name):
                with open(file_name, 'rb') as f:
                    files['file'] = (os.path.basename(file_name), f, 'application/zip')

                    for k, v in data.items():
                        files[k] = (None, v)

                    create_resp = client.request(request_url,
                                                 data=data,
                                                 files=files,
                                                 method=requests.post,
                                                 message_prefix='Dataset upload from zip file')

            # If not a zip, assert it is a CSV
            else:
                with open(file_name, 'r') as f:
                    files['file'] = (os.path.basename(file_name), f, 'text/csv')

                    for k, v in data.items():
                        files[k] = (None, v)

                    create_resp = client.request(request_url,
                                                 data=data,
                                                 files=files,
                                                 method=requests.post,
                                                 message_prefix='Dataset upload from csv file')

        if create_resp is None:
            raise PrevisionException('[Dataset] Unexpected case in dataset creation')

        create_json = parse_json(create_resp)
        url = '/{}/{}'.format(cls.resource, create_json['_id'])
        event_tuple = previsionio.utils.EventTuple('DATASET_UPDATE',
                                                   [('copy_state', 'done'),
                                                       ('describe_state', 'done'),
                                                       ('drift_state', 'done')],
                                                   [('copy_state', 'failed'),
                                                       ('describe_state', 'failed'),
                                                       ('drift_state', 'failed'),
                                                       ('embeddings_state', 'failed')])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(create_json['_id'],
                                                cls.resource,
                                                event_tuple,
                                                specific_url=url)

        dset_resp = client.request(url, method=requests.get, message_prefix='Dataset loading')
        dset_json = parse_json(dset_resp)

        if dataframe is not None and file_name is not None:
            os.remove(file_name)

        return cls(**dset_json)


class DatasetImages(ApiResource):

    """ DatasetImages objects represent image data resources that will be used by
        Prevision.io's platform.

        In order to launch an auto ml process (see :class:`.BaseUsecase` class), we need to have
        the matching dataset stored in the related workspace.

        Within the platform, image folder datasets are stored as ZIP files and are copied from
        ZIP files. """

    resource = 'image-folders'

    def __init__(self, _id: str, name: str, project_id: str, copy_state, **kwargs):
        super().__init__(_id=_id)
        self.name = name
        self._id = _id
        self.project_id = project_id
        self.copy_state = copy_state

        self.other_params = kwargs

    @classmethod
    def from_id(cls, _id: str) -> 'DatasetImages':
        return cls(**super()._from_id(_id=_id))

    def delete(self):
        """Delete a DatasetImages from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset images does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        resp = client.request(endpoint='/{}/{}'.format(self.resource, self.id),
                              method=requests.delete,
                              message_prefix='Image folder delete')
        return resp

    @classmethod
    def list(cls, project_id: str, all: bool = True):
        """ List all the available dataset image in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.DatasetImages` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.DatasetImages`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**conn_data) for conn_data in resources]

    @classmethod
    def _new(cls, project_id: str, name: str, file_name: str):
        """ Register a new image dataset in the workspace for further processing
        (in the image folders group).

        .. note::

            To start a new use case on a dataset image, it has to be already
            registred in your workspace.

        Args:
            name (str): Registration name for the dataset
            file_name (str): Path to the zip file to upload as image dataset

        Raises:
            PrevisionException: Error while creating the dataset on the platform

        Returns:
            :class:`.DatasetImages`: The registered dataset object in the current workspace.
        """
        request_url = '/projects/{}/{}'.format(project_id, cls.resource)
        source = open(file_name, 'rb')

        files = {
            'name': (None, name),
            'file': (name, source)
        }

        create_resp = client.request(request_url,
                                     files=files,
                                     method=requests.post,
                                     message_prefix='Image folder upload')
        source.close()

        create_json = parse_json(create_resp)
        url = '/{}/{}'.format(cls.resource, create_json['_id'])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(create_json['_id'],
                                                cls.resource,
                                                previsionio.utils.EventTuple(
                                                    'FOLDER_UPDATE'),
                                                specific_url=url)

        dset_resp = client.request(url, method=requests.get, message_prefix='Image folder loading')
        dset_json = parse_json(dset_resp)
        return cls(**dset_json)

    def download(self, download_path: str = None):
        """Download the dataset from the platform locally.

        Args:
            download_path (str, optional): Target local directory path
                (if none is provided, the current working directory is
                used)

        Returns:
            str: Path the data was downloaded to

        Raises:
            PrevisionException: If dataset does not exist or if there
                was another error fetching or parsing data
        """
        endpoint = '/{}/{}/download'.format(self.resource, self.id)
        resp = client.request(endpoint=endpoint,
                              method=requests.get,
                              message_prefix='Image folder download')
        if resp._content is None:
            raise PrevisionException('could not download dataset')
        if not download_path:
            download_path = os.getcwd()
        path = os.path.join(download_path, self.name + ".zip")
        with open(path, "wb") as file:
            file.write(resp._content)
        return path
